# sage_setup: distribution = sagemath-categories
"""
Multidimensional enumeration

AUTHORS:

- Joel B. Mohler (2006-10-12)

- William Stein (2006-07-19)

- Jon Hanke
"""

# *****************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
# *****************************************************************************
from sage.misc.misc_c import prod


def _len(L):
    """
    Determines the length of ``L``.

    Uses either ``cardinality`` or ``__len__`` as appropriate.

    EXAMPLES::

        sage: from sage.misc.mrange import _len
        sage: _len(ZZ)
        +Infinity
        sage: _len(range(4))
        4
        sage: _len(4)
        Traceback (most recent call last):
        ...
        TypeError: object of type 'sage.rings.integer.Integer' has no len()
    """
    try:
        return L.cardinality()
    except AttributeError:
        return len(L)


def _is_finite(L, fallback=True):
    """
    Determines whether ``L`` is finite.

    If ``L`` implements none of ``is_finite``, ``cardinality`` or
    ``__len__``, we assume by default that it is finite for speed
    reasons.  This choice can be overridden by passing in
    ``fallback``.

    EXAMPLES::

        sage: from sage.misc.mrange import _is_finite
        sage: _is_finite(ZZ)
        False
        sage: _is_finite(range(4))
        True
        sage: _is_finite([])
        True
        sage: _is_finite(range(10^8))
        True
        sage: from itertools import product
        sage: _is_finite(product([1],[1])) # does not provide is_finite() or __len__()
        True
    """
    try:
        return L.is_finite()
    except AttributeError:
        pass
    except ValueError:
        # L doesn't know how to determine whether it's finite
        return fallback

    try:
        n = _len(L)
    except (TypeError, AttributeError, NotImplementedError):
        # We usually assume L is finite for speed reasons
        return fallback

    from sage.rings.infinity import infinity
    if n is infinity:
        return False
    return True


def _xmrange_iter(iter_list, typ=list):
    """
    This implements the logic for :func:`mrange_iter` and :class:`xmrange_iter`.

    Note that with typ==list, we will be returning a new copy each
    iteration. This makes it OK to modified the returned list. This
    functionality is relied on in the polynomial iterators. Here's a
    doc-test to prove this::

        sage: iter = sage.misc.mrange._xmrange_iter( [[1,2],[1,3]] )
        sage: l1 = next(iter)
        sage: l2 = next(iter)
        sage: l1 is l2
        False

    However, if you would like to reuse the list object::

        sage: iter = sage.misc.mrange._xmrange_iter( [[1,2],[1,3]], lambda x: x )
        sage: l1 = next(iter)
        sage: l2 = next(iter)
        sage: l1 is l2  # eeek, this is freaky!
        True

    We check that :issue:`14285` has been resolved::

        sage: iter = sage.misc.mrange._xmrange_iter([ZZ,[]])
        sage: next(iter)
        Traceback (most recent call last):
        ...
        StopIteration

    We check that :issue:`28521` is fixed::

        sage: next(sage.misc.mrange._xmrange_iter([[], [1]]))
        Traceback (most recent call last):
        ...
        StopIteration
        sage: next(sage.misc.mrange._xmrange_iter([[1], []]))
        Traceback (most recent call last):
        ...
        StopIteration
        sage: next(sage.misc.mrange._xmrange_iter([[1], [], [2]]))
        Traceback (most recent call last):
        ...
        StopIteration
    """
    if len(iter_list) == 0:
        yield typ()
        return
    # If any iterator in the list is infinite we need to be more careful
    if any(not _is_finite(L) for L in iter_list):
        for L in iter_list:
            try:
                n = _len(L)
            except TypeError:
                continue
        if n == 0:
            return
    curr_iters = [iter(i) for i in iter_list]
    try:
        curr_elt = [next(i) for i in curr_iters[:-1]]
    except StopIteration:
        return
    curr_elt.append(None)
    place = len(iter_list) - 1
    while True:
        try:
            while True:
                curr_elt[place] = next(curr_iters[place])
                if place < len(iter_list) - 1:
                    place += 1
                    curr_iters[place] = iter(iter_list[place])
                    continue
                else:
                    yield typ(curr_elt)
        except StopIteration:
            place -= 1
            if place == -1:
                return


def mrange_iter(iter_list, typ=list):
    """
    Return the multirange list derived from the given list of iterators.

    This is the list version of :func:`xmrange_iter`. Use :class:`xmrange_iter`
    for the iterator.

    More precisely, return the iterator over all objects of type ``typ`` of
    n-tuples of Python ints with entries between 0 and the integers in
    the sizes list. The iterator is empty if sizes is empty or contains
    any nonpositive integer.

    INPUT:

    - ``iter_list`` -- a finite iterable of finite iterables

    - ``typ`` -- (default: list) a type or class; more
      generally, something that can be called with a list as input

    OUTPUT: list

    EXAMPLES::

        sage: mrange_iter([range(3),[0,2]])
        [[0, 0], [0, 2], [1, 0], [1, 2], [2, 0], [2, 2]]
        sage: mrange_iter([['Monty','Flying'],['Python','Circus']], tuple)
        [('Monty', 'Python'), ('Monty', 'Circus'), ('Flying', 'Python'), ('Flying', 'Circus')]
        sage: mrange_iter([[2,3,5,7],[1,2]], sum)
        [3, 4, 4, 5, 6, 7, 8, 9]

    Examples that illustrate empty multi-ranges::

        sage: mrange_iter([range(5),range(3),range(0)])
        []
        sage: mrange_iter([range(5), range(3), range(-2)])
        []

    This example is not empty, and should not be. See :issue:`6561`.

    ::

        sage: mrange_iter([])
        [[]]

    AUTHORS:

    - Joel B. Mohler
    """
    return list(_xmrange_iter(iter_list, typ))


class xmrange_iter:
    """
    Return the multirange iterate derived from the given iterators and
    type.

    .. NOTE::

       This basically gives you the Cartesian product of sets.

    More precisely, return the iterator over all objects of type typ of
    n-tuples of Python ints with entries between 0 and the integers in
    the sizes list. The iterator is empty if sizes is empty or contains
    any nonpositive integer.

    Use :func:`mrange_iter` for the non-iterator form.

    INPUT:

    - ``iter_list`` -- list of objects usable as iterators (possibly
      lists)

    - ``typ`` -- (default: list) a type or class; more generally,
      something that can be called with a list as input

    OUTPUT: a generator

    EXAMPLES: We create multi-range iterators, print them and also
    iterate through a tuple version. ::

        sage: z = xmrange_iter([list(range(3)),list(range(2))], tuple);z
        xmrange_iter([[0, 1, 2], [0, 1]], <... 'tuple'>)
        sage: for a in z:
        ....:     print(a)
        (0, 0)
        (0, 1)
        (1, 0)
        (1, 1)
        (2, 0)
        (2, 1)

    We illustrate a few more iterations.

    ::

        sage: list(xmrange_iter([range(3),range(2)]))
        [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]]
        sage: list(xmrange_iter([range(3),range(2)], tuple))
        [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)]

    Here we compute the sum of each element of the multi-range
    iterator::

        sage: list(xmrange_iter([range(3),range(2)], sum))
        [0, 1, 1, 2, 2, 3]

    Next we compute the product::

        sage: list(xmrange_iter([range(3),range(2)], prod))
        [0, 0, 0, 1, 0, 2]

    Examples that illustrate empty multi-ranges.

    ::

        sage: list(xmrange_iter([range(5),range(3),range(-2)]))
        []
        sage: list(xmrange_iter([range(5),range(3),range(0)]))
        []

    This example is not empty, and should not be. See :issue:`6561`.

    ::

        sage: list(xmrange_iter([]))
        [[]]

    We use a multi-range iterator to iterate through the Cartesian
    product of sets.

    ::

        sage: X = ['red', 'apple', 389]
        sage: Y = ['orange', 'horse']
        sage: for i,j in xmrange_iter([X, Y], tuple):
        ....:     print((i, j))
        ('red', 'orange')
        ('red', 'horse')
        ('apple', 'orange')
        ('apple', 'horse')
        (389, 'orange')
        (389, 'horse')

    AUTHORS:

    - Joel B. Mohler
    """
    def __init__(self, iter_list, typ=list):
        self.iter_list = iter_list
        self.typ = typ

    def __repr__(self):
        if self.typ == list:
            return 'xmrange_iter(%s)' % self.iter_list
        else:
            return 'xmrange_iter(%s, %s)' % (self.iter_list, self.typ)

    def __iter__(self):
        return _xmrange_iter(self.iter_list, self.typ)

    def __len__(self):
        """
        Return the cardinality of this iterator as an int.

        This raises a :exc:`TypeError` if the cardinality does not fit
        into a Python int.

        EXAMPLES::

            sage: C = cartesian_product_iterator([range(3),range(4)])
            sage: len(C)
            12
            sage: len(cartesian_product_iterator([]))
            1
            sage: len(cartesian_product_iterator([ZZ,[]]))
            0
            sage: len(cartesian_product_iterator([ZZ,ZZ]))
            Traceback (most recent call last):
            ...
            TypeError: cardinality does not fit into a Python int
        """
        n = self.cardinality()
        try:
            n = int(n)
        except TypeError:
            raise TypeError("cardinality does not fit into a Python int")
        return n

    def cardinality(self):
        """
        Return the cardinality of this iterator.

        EXAMPLES::

            sage: C = cartesian_product_iterator([range(3),range(4)])
            sage: C.cardinality()
            12
            sage: C = cartesian_product_iterator([ZZ,QQ])
            sage: C.cardinality()
            +Infinity
            sage: C = cartesian_product_iterator([ZZ,[]])
            sage: C.cardinality()
            0
        """
        from sage.rings.integer import Integer
        from sage.rings.infinity import infinity
        ans = Integer(1)
        found_infinity = False
        for L in self.iter_list:
            try:
                n = L.cardinality()
            except AttributeError:
                n = Integer(len(L))
            if n == 0:
                return Integer(0)
            elif n is infinity:
                found_infinity = True
            elif not found_infinity:
                ans *= n
        if found_infinity:
            return infinity
        else:
            return ans


def _xmrange(sizes, typ=list):
    n = len(sizes)
    if n == 0:
        yield typ([])
        return
    for i in sizes:
        if i <= 0:
            return
    v = [0] * n    # make a list of n 0's.
    v[-1] = -1
    ptr_max = n - 1
    ptr = ptr_max
    while True:
        while True:
            if ptr != -1 and v[ptr] + 1 < sizes[ptr]:
                v[ptr] += 1
                ptr = ptr_max
                break
            elif ptr != -1:
                v[ptr] = 0
                ptr -= 1
            else:
                return
        yield typ(v)   # make a copy of v!


def mrange(sizes, typ=list):
    """
    Return the multirange list with given sizes and type.

    This is the list version of :class:`xmrange`.
    Use :class:`xmrange` for the iterator.

    More precisely, return the iterator over all objects of type typ of
    n-tuples of Python ints with entries between 0 and the integers in
    the sizes list. The iterator is empty if sizes is empty or contains
    any nonpositive integer.

    INPUT:

    - ``sizes`` -- list of nonnegative integers

    - ``typ`` -- (default: list) a type or class; more
      generally, something that can be called with a list as input

    OUTPUT: list

    EXAMPLES::

        sage: mrange([3,2])
        [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]]
        sage: mrange([3,2], tuple)
        [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)]
        sage: mrange([3,2], sum)
        [0, 1, 1, 2, 2, 3]

    Examples that illustrate empty multi-ranges::

        sage: mrange([5,3,-2])
        []
        sage: mrange([5,3,0])
        []

    This example is not empty, and should not be. See :issue:`6561`.

    ::

        sage: mrange([])
        [[]]

    AUTHORS:

    - Jon Hanke

    - William Stein
    """
    return list(_xmrange(sizes, typ))


class xmrange:
    """
    Return the multirange iterate with given sizes and type.

    More precisely, return the iterator over all objects of type typ of
    n-tuples of Python ints with entries between 0 and the integers in
    the sizes list. The iterator is empty if sizes is empty or contains
    any nonpositive integer.

    Use mrange for the non-iterator form.

    INPUT:

    - ``sizes`` -- list of nonnegative integers

    - ``typ`` -- (default: list) a type or class; more
      generally, something that can be called with a list as input

    OUTPUT: a generator

    EXAMPLES: We create multi-range iterators, print them and also
    iterate through a tuple version.

    ::

        sage: z = xmrange([3,2]);z
        xmrange([3, 2])
        sage: z = xmrange([3,2], tuple);z
        xmrange([3, 2], <... 'tuple'>)
        sage: for a in z:
        ....:     print(a)
        (0, 0)
        (0, 1)
        (1, 0)
        (1, 1)
        (2, 0)
        (2, 1)

    We illustrate a few more iterations.

    ::

        sage: list(xmrange([3,2]))
        [[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]]
        sage: list(xmrange([3,2], tuple))
        [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)]

    Here we compute the sum of each element of the multi-range
    iterator::

        sage: list(xmrange([3,2], sum))
        [0, 1, 1, 2, 2, 3]

    Next we compute the product::

        sage: list(xmrange([3,2], prod))
        [0, 0, 0, 1, 0, 2]

    Examples that illustrate empty multi-ranges.

    ::

        sage: list(xmrange([5,3,-2]))
        []
        sage: list(xmrange([5,3,0]))
        []

    This example is not empty, and should not be. See :issue:`6561`.

    ::

        sage: list(xmrange([]))
        [[]]

    We use a multi-range iterator to iterate through the Cartesian
    product of sets.

    ::

        sage: X = ['red', 'apple', 389]
        sage: Y = ['orange', 'horse']
        sage: for i,j in xmrange([len(X), len(Y)]):
        ....:     print((X[i], Y[j]))
        ('red', 'orange')
        ('red', 'horse')
        ('apple', 'orange')
        ('apple', 'horse')
        (389, 'orange')
        (389, 'horse')

    AUTHORS:

    - Jon Hanke

    - William Stein
    """
    def __init__(self, sizes, typ=list):
        self.sizes = [int(x) for x in sizes]
        self.typ = typ

    def __repr__(self):
        if self.typ == list:
            return 'xmrange(%s)' % self.sizes
        else:
            return 'xmrange(%s, %s)' % (self.sizes, self.typ)

    def __len__(self):
        sizes = self.sizes
        n = len(sizes)
        if n == 0:
            return 0
        for i in sizes:
            if i <= 0:
                return 0
        return prod(sizes, 1)

    def __iter__(self):
        return _xmrange(self.sizes, self.typ)


def cartesian_product_iterator(X):
    """
    Iterate over the Cartesian product.

    INPUT:

    - ``X`` -- list or tuple of lists

    OUTPUT: iterator over the Cartesian product of the elements of X

    EXAMPLES::

        sage: list(cartesian_product_iterator([[1,2], ['a','b']]))
        [(1, 'a'), (1, 'b'), (2, 'a'), (2, 'b')]
        sage: list(cartesian_product_iterator([]))
        [()]

    TESTS:

    Check that :issue:`28521` is fixed::

        sage: list(cartesian_product_iterator([[], [1]]))
        []
        sage: list(cartesian_product_iterator([[1], []]))
        []
        sage: list(cartesian_product_iterator([[1], [], [2]]))
        []
    """
    return xmrange_iter(X, tuple)


def cantor_product(*args, **kwds):
    r"""
    Return an iterator over the product of the inputs along the diagonals a la
    :wikipedia:`Cantor pairing <Pairing_function#Cantor_pairing_function>`.

    INPUT:

    - a certain number of iterables

    - ``repeat`` -- an optional integer. If it is provided, the input is
      repeated ``repeat`` times

    Other keyword arguments are passed to
    :class:`sage.combinat.integer_lists.invlex.IntegerListsLex`.

    EXAMPLES::

        sage: from sage.misc.mrange import cantor_product
        sage: list(cantor_product([0, 1], repeat=3))
        [(0, 0, 0),
         (1, 0, 0),
         (0, 1, 0),
         (0, 0, 1),
         (1, 1, 0),
         (1, 0, 1),
         (0, 1, 1),
         (1, 1, 1)]
        sage: list(cantor_product([0, 1], [0, 1, 2, 3]))
        [(0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2), (0, 3), (1, 3)]

    Infinite iterators are valid input as well::

       sage: from itertools import islice
       sage: list(islice(cantor_product(ZZ, QQ), 14r))
        [(0, 0),
         (1, 0),
         (0, 1),
         (-1, 0),
         (1, 1),
         (0, -1),
         (2, 0),
         (-1, 1),
         (1, -1),
         (0, 1/2),
         (-2, 0),
         (2, 1),
         (-1, -1),
         (1, 1/2)]

    TESTS::

        sage: C = cantor_product([0, 1], [0, 1, 2, 3], [0, 1, 2])
        sage: sum(1 for _ in C) == 2*4*3
        True

        sage: from itertools import count
        sage: list(cantor_product([], count()))
        []
        sage: list(cantor_product(count(), [], count()))
        []

        sage: list(cantor_product(count(), repeat=0))
        [()]

        sage: next(cantor_product(count(), repeat=-1))
        Traceback (most recent call last):
        ...
        ValueError: repeat argument cannot be negative
        sage: next(cantor_product(count(), toto='hey'))
        Traceback (most recent call last):
        ...
        TypeError: ...__init__() got an unexpected keyword argument 'toto'

    ::

        sage: list(cantor_product(srange(5), repeat=2, min_slope=1))
        [(0, 1), (0, 2), (1, 2), (0, 3), (1, 3),
         (0, 4), (2, 3), (1, 4), (2, 4), (3, 4)]

    Check that :issue:`24897` is fixed::

        sage: from sage.misc.mrange import cantor_product
        sage: list(cantor_product([1]))
        [(1,)]
        sage: list(cantor_product([1], repeat=2))
        [(1, 1)]
        sage: list(cantor_product([1], [1,2]))
        [(1, 1), (1, 2)]
        sage: list(cantor_product([1,2], [1]))
        [(1, 1), (2, 1)]
    """
    from itertools import count
    from sage.combinat.integer_lists import IntegerListsLex

    m = len(args)                         # numer of factors
    lengths = [None] * m                  # None or length of factors
    data = [[] for _ in range(m)]         # the initial slice of each factor
    iterators = [iter(a) for a in args]   # the iterators
    repeat = int(kwds.pop('repeat', 1))
    if repeat == 0:
        yield ()
        return
    elif repeat < 0:
        raise ValueError("repeat argument cannot be negative")
    mm = m * repeat

    for n in count(0):
        # try to add one more term to each bin
        for i, a in enumerate(iterators):
            if lengths[i] is None:
                try:
                    data[i].append(next(a))
                except StopIteration:
                    assert len(data[i]) == n
                    if n == 0:
                        return
                    lengths[i] = n

        # iterate through what we have
        ceiling = [n if lengths[i] is None else lengths[i] - 1
                   for i in range(m)] * repeat
        for v in IntegerListsLex(n, length=mm, ceiling=ceiling, **kwds):
            yield tuple(data[i % m][v[i]] for i in range(mm))

        if all(l is not None for l in lengths) and repeat * sum(l - 1 for l in lengths) <= n:
            return



# sage_setup: distribution = sagemath-categories
r"""
Enumeration of rational points on projective schemes

Naive algorithms for enumerating rational points over `\QQ` or finite fields
over for general schemes.

.. WARNING::

    Incorrect results and infinite loops may occur if using a wrong function.
    (For instance using an affine function for a projective scheme or a finite
    field function for a scheme defined over an infinite field.)

EXAMPLES:

Projective, over `\QQ`::

    sage: from sage.schemes.projective.projective_rational_point import enum_projective_rational_field
    sage: P.<X,Y,Z> = ProjectiveSpace(2, QQ)
    sage: C = P.subscheme([X + Y - Z])
    sage: enum_projective_rational_field(C, 3)
    [(-2 : 3 : 1), (-1 : 1 : 0), (-1 : 2 : 1), (-1/2 : 3/2 : 1),
     (0 : 1 : 1), (1/3 : 2/3 : 1), (1/2 : 1/2 : 1), (2/3 : 1/3 : 1),
     (1 : 0 : 1), (3/2 : -1/2 : 1), (2 : -1 : 1), (3 : -2 : 1)]


Projective over a finite field::

    sage: from sage.schemes.projective.projective_rational_point import enum_projective_finite_field
    sage: E = EllipticCurve('72').change_ring(GF(19))                                   # needs sage.schemes
    sage: enum_projective_finite_field(E)                                               # needs sage.schemes
    [(0 : 1 : 0), (1 : 0 : 1), (3 : 0 : 1), (4 : 9 : 1), (4 : 10 : 1),
     (6 : 6 : 1), (6 : 13 : 1), (7 : 6 : 1), (7 : 13 : 1), (9 : 4 : 1),
     (9 : 15 : 1), (12 : 8 : 1), (12 : 11 : 1), (13 : 8 : 1), (13 : 11 : 1),
     (14 : 3 : 1), (14 : 16 : 1), (15 : 0 : 1), (16 : 9 : 1), (16 : 10 : 1),
     (17 : 7 : 1), (17 : 12 : 1), (18 : 9 : 1), (18 : 10 : 1)]


AUTHORS:

- David R. Kohel <kohel@maths.usyd.edu.au>: original version.

- John Cremona and Charlie Turner <charlotteturner@gmail.com> (06-2010):
  improvements to clarity and documentation.

- Raghukul Raman <raghukul.raman01@gmail.com> (2018): Added sieve algorithm
"""

# ****************************************************************************
#       Copyright (C) 2010 William Stein, David Kohel, John Cremona, Charlie Turner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from itertools import product

from sage.arith.misc import gcd, next_prime, previous_prime, crt
from sage.arith.srange import srange
from sage.misc.lazy_import import lazy_import
from sage.misc.misc_c import prod
from sage.misc.mrange import xmrange
from sage.schemes.generic.scheme import Scheme
from sage.parallel.ncpus import ncpus
from sage.parallel.use_fork import p_iter_fork
from sage.rings.finite_rings.finite_field_constructor import FiniteField as GF
from sage.rings.integer_ring import ZZ
from sage.schemes.generic.scheme import is_Scheme

lazy_import('sage.matrix.constructor', 'matrix')
lazy_import('sage.rings.real_mpfr', 'RR')


def enum_projective_rational_field(X, B):
    r"""
    Enumerates projective, rational points on scheme ``X`` of height up to
    bound ``B``.

    INPUT:

    - ``X`` -- a scheme or set of abstract rational points of a scheme

    - ``B`` -- a positive integer bound

    OUTPUT:

    A list containing the projective points of ``X`` of height up to ``B``,
    sorted.

    EXAMPLES::

        sage: P.<X,Y,Z> = ProjectiveSpace(2, QQ)
        sage: C = P.subscheme([X + Y - Z])
        sage: from sage.schemes.projective.projective_rational_point import enum_projective_rational_field
        sage: enum_projective_rational_field(C(QQ), 6)
        [(-5 : 6 : 1), (-4 : 5 : 1), (-3 : 4 : 1), (-2 : 3 : 1),
         (-3/2 : 5/2 : 1), (-1 : 1 : 0), (-1 : 2 : 1), (-2/3 : 5/3 : 1),
         (-1/2 : 3/2 : 1), (-1/3 : 4/3 : 1), (-1/4 : 5/4 : 1),
         (-1/5 : 6/5 : 1), (0 : 1 : 1), (1/6 : 5/6 : 1), (1/5 : 4/5 : 1),
         (1/4 : 3/4 : 1), (1/3 : 2/3 : 1), (2/5 : 3/5 : 1), (1/2 : 1/2 : 1),
         (3/5 : 2/5 : 1), (2/3 : 1/3 : 1), (3/4 : 1/4 : 1), (4/5 : 1/5 : 1),
         (5/6 : 1/6 : 1), (1 : 0 : 1), (6/5 : -1/5 : 1), (5/4 : -1/4 : 1),
         (4/3 : -1/3 : 1), (3/2 : -1/2 : 1), (5/3 : -2/3 : 1), (2 : -1 : 1),
         (5/2 : -3/2 : 1), (3 : -2 : 1), (4 : -3 : 1), (5 : -4 : 1),
         (6 : -5 : 1)]
        sage: enum_projective_rational_field(C,6) == enum_projective_rational_field(C(QQ),6)
        True

    ::

        sage: P3.<W,X,Y,Z> = ProjectiveSpace(3, QQ)
        sage: enum_projective_rational_field(P3, 1)
        [(-1 : -1 : -1 : 1), (-1 : -1 : 0 : 1), (-1 : -1 : 1 : 0), (-1 : -1 : 1 : 1),
         (-1 : 0 : -1 : 1), (-1 : 0 : 0 : 1), (-1 : 0 : 1 : 0), (-1 : 0 : 1 : 1),
         (-1 : 1 : -1 : 1), (-1 : 1 : 0 : 0), (-1 : 1 : 0 : 1), (-1 : 1 : 1 : 0),
         (-1 : 1 : 1 : 1), (0 : -1 : -1 : 1), (0 : -1 : 0 : 1), (0 : -1 : 1 : 0),
         (0 : -1 : 1 : 1), (0 : 0 : -1 : 1), (0 : 0 : 0 : 1), (0 : 0 : 1 : 0),
         (0 : 0 : 1 : 1), (0 : 1 : -1 : 1), (0 : 1 : 0 : 0), (0 : 1 : 0 : 1),
         (0 : 1 : 1 : 0), (0 : 1 : 1 : 1), (1 : -1 : -1 : 1), (1 : -1 : 0 : 1),
         (1 : -1 : 1 : 0), (1 : -1 : 1 : 1), (1 : 0 : -1 : 1), (1 : 0 : 0 : 0),
         (1 : 0 : 0 : 1), (1 : 0 : 1 : 0), (1 : 0 : 1 : 1), (1 : 1 : -1 : 1),
         (1 : 1 : 0 : 0), (1 : 1 : 0 : 1), (1 : 1 : 1 : 0), (1 : 1 : 1 : 1)]

    ALGORITHM:

    We just check all possible projective points in correct dimension
    of projective space to see if they lie on ``X``.

    AUTHORS:

    - John Cremona and Charlie Turner (06-2010)
    """
    from sage.schemes.projective.projective_space import ProjectiveSpace_ring
    if isinstance(X, Scheme):
        if not isinstance(X.ambient_space(), ProjectiveSpace_ring):
            raise TypeError("ambient space must be projective space over the rational field")
        X = X(X.base_ring())
    elif not isinstance(X.codomain().ambient_space(), ProjectiveSpace_ring):
        raise TypeError("codomain must be projective space over the rational field")

    n = X.codomain().ambient_space().ngens()
    zero = (0,) * n
    pts = []
    for c in product(*[srange(-B, B + 1) for _ in range(n)]):
        if gcd(c) == 1 and c > zero:
            try:
                pts.append(X(c))
            except TypeError:
                pass
    pts.sort()
    return pts


def enum_projective_number_field(X, **kwds):
    """
    Enumerates projective points on scheme ``X`` defined over a number field.

    Simply checks all of the points of absolute height of at most ``B``
    and adds those that are on the scheme to the list.

    This algorithm computes 2 lists: L containing elements x in `K` such that
    H_k(x) <= B, and a list L' containing elements x in `K` that, due to
    floating point issues,
    may be slightly larger then the bound. This can be controlled
    by lowering the tolerance.

    ALGORITHM:

    This is an implementation of the revised algorithm (Algorithm 4) in
    [DK2013]_. Algorithm 5 is used for imaginary quadratic fields.

    INPUT: keyword arguments:

    - ``bound`` -- a real number

    - ``tolerance`` -- a rational number in (0,1] used in Doyle-Krumm
      algorithm-4

    - ``precision`` -- the precision to use for computing the elements of
      bounded height of number fields

    OUTPUT: a sorted list containing the projective points of ``X`` of absolute
    height up to ``B``

    EXAMPLES::

        sage: # needs sage.rings.number_field
        sage: from sage.schemes.projective.projective_rational_point import enum_projective_number_field
        sage: u = QQ['u'].0
        sage: K = NumberField(u^3 - 5, 'v')
        sage: P.<x,y,z> = ProjectiveSpace(K, 2)
        sage: X = P.subscheme([x - y])
        sage: enum_projective_number_field(X(K), bound=RR(5^(1/3)), prec=2^10)          # needs sage.symbolic
        [(0 : 0 : 1), (1 : 1 : 0), (-1 : -1 : 1), (1 : 1 : 1)]

    ::

        sage: # needs sage.rings.number_field
        sage: u = QQ['u'].0
        sage: K = NumberField(u^2 + 3, 'v')
        sage: A.<x,y> = ProjectiveSpace(K, 1)
        sage: X = A.subscheme(x - y)
        sage: enum_projective_number_field(X, bound=2)
        [(1 : 1)]
    """
    B = kwds.pop('bound')
    tol = kwds.pop('tolerance', 1e-2)
    prec = kwds.pop('precision', 53)
    from sage.schemes.projective.projective_space import ProjectiveSpace_ring
    if isinstance(X, Scheme):
        if not isinstance(X.ambient_space(), ProjectiveSpace_ring):
            raise TypeError("ambient space must be projective space over a number field")
        X = X(X.base_ring())
    else:
        if not isinstance(X.codomain().ambient_space(), ProjectiveSpace_ring):
            raise TypeError("codomain must be projective space over a number field")

    R = X.codomain().ambient_space()

    pts = []

    for P in R.points_of_bounded_height(bound=B, tolerance=tol, precision=prec):
        try:
            pts.append(X(P))
        except TypeError:
            pass
    pts.sort()
    return pts


def enum_projective_finite_field(X):
    """
    Enumerates projective points on scheme ``X`` defined over a finite field.

    INPUT:

    - ``X`` -- a scheme defined over a finite field or a set of abstract
      rational points of such a scheme

    OUTPUT:

    A list containing the projective points of ``X`` over the finite field,
    sorted.

    EXAMPLES::

        sage: from sage.schemes.projective.projective_rational_point import enum_projective_finite_field
        sage: F = GF(53)
        sage: P.<X,Y,Z> = ProjectiveSpace(2, F)
        sage: len(enum_projective_finite_field(P(F)))
        2863
        sage: 53^2 + 53 + 1
        2863

    ::

        sage: # needs sage.rings.finite_rings
        sage: F = GF(9, 'a')
        sage: P.<X,Y,Z> = ProjectiveSpace(2,F)
        sage: C = Curve(X^3 - Y^3 + Z^2*Y)                                              # needs sage.schemes
        sage: enum_projective_finite_field(C(F))                                        # needs sage.schemes
        [(0 : 0 : 1), (0 : 1 : 1), (0 : 2 : 1), (1 : 1 : 0), (a + 1 : 2*a : 1),
         (a + 1 : 2*a + 1 : 1), (a + 1 : 2*a + 2 : 1), (2*a + 2 : a : 1),
         (2*a + 2 : a + 1 : 1), (2*a + 2 : a + 2 : 1)]

    ::

        sage: F = GF(5)
        sage: P2F.<X,Y,Z> = ProjectiveSpace(2, F)
        sage: enum_projective_finite_field(P2F)
        [(0 : 0 : 1), (0 : 1 : 0), (0 : 1 : 1), (0 : 2 : 1), (0 : 3 : 1), (0 : 4 : 1),
         (1 : 0 : 0), (1 : 0 : 1), (1 : 1 : 0), (1 : 1 : 1), (1 : 2 : 1), (1 : 3 : 1),
         (1 : 4 : 1), (2 : 0 : 1), (2 : 1 : 0), (2 : 1 : 1), (2 : 2 : 1), (2 : 3 : 1),
         (2 : 4 : 1), (3 : 0 : 1), (3 : 1 : 0), (3 : 1 : 1), (3 : 2 : 1), (3 : 3 : 1),
         (3 : 4 : 1), (4 : 0 : 1), (4 : 1 : 0), (4 : 1 : 1), (4 : 2 : 1), (4 : 3 : 1),
         (4 : 4 : 1)]

    ALGORITHM:

    Checks all points in projective space to see if they lie on ``X``.

    .. WARNING::

        If ``X`` is defined over an infinite field, this code will not finish!

    AUTHORS:

    - John Cremona and Charlie Turner (06-2010).
    """
    from sage.schemes.projective.projective_space import ProjectiveSpace_ring
    if isinstance(X, Scheme):
        if not isinstance(X.ambient_space(), ProjectiveSpace_ring):
            raise TypeError("ambient space must be projective space over a finite")
        X = X(X.base_ring())
    elif not isinstance(X.codomain().ambient_space(), ProjectiveSpace_ring):
        raise TypeError("codomain must be projective space over a finite field")

    n = X.codomain().ambient_space().ngens() - 1
    F = X.value_ring()
    pts = []
    for k in range(n + 1):
        for c in product(*[F for _ in range(k)]):
            try:
                pts.append(X(list(c) + [1] + [0] * (n - k)))
            except TypeError:
                pass
    pts.sort()
    return pts


def sieve(X, bound):
    r"""
    Return the list of all projective, rational points on scheme ``X`` of
    height up to ``bound``.

    Height of a projective point `X = (x_1, x_2,\dots, x_n)` is given by
    `H_X = \max(y_1, y_2,\dots, y_n)`, where the values `y_i`
    are the normalized coordinates such that all `y_i` are integers and
    `\gcd(y_1, y_2,\dots, y_n) = 1`.

    ALGORITHM:

    Main idea behind the algorithm is to find points modulo primes
    and then reconstruct them using chinese remainder theorem.
    We find modulo primes parallelly and then lift them and apply
    LLL in parallel.

    For the algorithm to work correctly, sufficient primes need
    to be present, these are calculated using the bound given in
    this([Hutz2015]_) paper.

    INPUT:

    - ``X`` -- a scheme with ambient space defined over projective space

    - ``bound`` -- positive integer bound

    OUTPUT:

    A list containing the projective rational points of ``X`` of height
    up to ``bound``, sorted

    EXAMPLES::

        sage: from sage.schemes.projective.projective_rational_point import sieve
        sage: P.<x,y,z,q> = ProjectiveSpace(QQ, 3)
        sage: Y = P.subscheme([x^2 - 3^2*y^2 + z*q, x + z + 4*q])
        sage: sorted(sieve(Y, 12))              # long time                             # needs sage.libs.singular
        [(-4 : -4/3 : 0 : 1), (-4 : 4/3 : 0 : 1),
         (-1 : -1/3 : 1 : 0), (-1 : 1/3 : 1 : 0)]

    ::

        sage: from sage.schemes.projective.projective_rational_point import sieve
        sage: E = EllipticCurve('37a')                                                  # needs sage.schemes
        sage: sorted(sieve(E, 14))              # long time                             # needs sage.libs.singular sage.schemes
        [(0 : 1 : 0),
         (-1 : -1 : 1),
         (-1 : 0 : 1),
         (0 : -1 : 1),
         (0 : 0 : 1),
         (1/4 : -5/8 : 1),
         (1/4 : -3/8 : 1),
         (1 : -1 : 1),
         (1 : 0 : 1),
         (2 : -3 : 1),
         (2 : 2 : 1),
         (6 : 14 : 1)]

    TESTS:

    Algorithm works even if coefficients are fraction::

        sage: from sage.schemes.projective.projective_rational_point import sieve
        sage: P.<x,y,z> = ProjectiveSpace(2, QQ)
        sage: X = P.subscheme(3*x - 3/2*y)
        sage: sieve(X, 3)                                                               # needs sage.libs.singular
        [(-1 : -2 : 1), (-1/2 : -1 : 1), (-1/3 : -2/3 : 1), (0 : 0 : 1),
         (1/3 : 2/3 : 1), (1/2 : 1 : 0), (1/2 : 1 : 1), (1 : 2 : 1)]
    """
    if bound < 1: # no projective rational point with height less than 1
        return []

    modulo_points = [] # list to store point modulo primes
    len_modulo_points = [] # stores number of points with respect to each prime
    primes_list = [] # list of good primes

    X.normalize_defining_polynomials()

    P = X.ambient_space()
    N = P.dimension()
    dim_scheme = X.dimension()

    # bound as per preposition - 4, in preperiodic points paper
    B = RR(2**(N/4+1)*bound**2*(N+1).sqrt())

    m = [0 for _ in range(N + 1)]

    def sufficient_primes(x):
        r"""
        Return a list of primes whose product is > `x`.
        """
        small_primes = [2,3]
        prod_primes = 6

        while prod_primes < x:
            p = next_prime(small_primes[-1])
            small_primes.append(p)
            prod_primes *= p
        return small_primes

    def good_primes(B):
        r"""
        Given the bound returns the prime whose product is greater than ``B``
        and which would take least amount of time to run main sieve algorithm

        Complexity of finding points modulo primes is assumed to be N^2 * P_max^{N}.
        Complexity of lifting points and LLL() function is assumed to
        be close to N^5 * (alpha^dim_scheme / P_max).
        where alpha is product of all primes, and P_max is largest prime in list.
        """

        M = {}  # stores optimal list of primes, corresponding to list size
        small_primes = sufficient_primes(B)
        max_length = len(small_primes)
        M[max_length] = small_primes
        current_count = max_length - 1

        while current_count > 1:
            current_list = [] # stores prime which are bigger than least
            updated_list = []
            best_list = []

            least = (RR(B)**(1.00/current_count)).floor()
            for i in range(current_count):
                current_list.append(next_prime(least))
                least = current_list[-1]
            # improving list of primes by taking prime less than least
            # this part of algorithm is used to centralize primes around `least`
            prod_prime = prod(current_list)
            least = current_list[0]
            while least != 2 and prod_prime > B and len(updated_list) < current_count:
                best_list = updated_list + current_list[:current_count - len(updated_list)]
                updated_list.append(previous_prime(least))
                least = updated_list[-1]

                removed_prime = current_list[current_count - len(updated_list)]
                prod_prime = (prod_prime * least) / removed_prime

            M[current_count] = sorted(best_list)
            current_count = current_count - 1

        best_size = 2
        best_time = (N**2)*M[2][-1]**(N) + (N**5 * RR(prod(M[2])**dim_scheme / M[2][-1]) )
        for i in range(2, max_length + 1):
            current_time = (N**2)*M[i][-1]**(N) + (N**5 * RR(prod(M[i])**dim_scheme / M[i][-1]) )
            if current_time < best_time:
                best_size = i
                best_time = current_time

        return M[best_size]

    def parallel_function(X, p):
        r"""
        Function used in parallel computation, computes a list of
        all rational points in modulo ring.
        """
        Xp = X.change_ring(GF(p))
        L = Xp.rational_points()

        return [list(_) for _ in L]

    def points_modulo_primes(X, primes):
        r"""
        Return a list of rational points modulo all `p` in primes,
        computed parallelly.
        """
        normalized_input = []
        for p in primes_list:
            normalized_input.append(((X, p, ), {}))
        p_iter = p_iter_fork(ncpus())

        points_pair = list(p_iter(parallel_function, normalized_input))
        points_pair.sort()
        return [pair[1] for pair in points_pair]

    def parallel_function_combination(point_p_max):
        r"""
        Function used in parallel computation, computes rational
        points lifted.
        """
        rat_points = set()
        for tupl in xmrange(len_modulo_points):
            point = []
            for k in range(N + 1):
                # lift all coordinates of given point using chinese remainder theorem
                L = [modulo_points[j][tupl[j]][k].lift() for j in range(len_primes - 1)]
                L.append(point_p_max[k].lift())
                point.append( crt(L, primes_list) )

            for i in range(N+1):
                m[i] = point[i]

            M = matrix(ZZ, N+2, N+1, m)
            A = M.LLL()
            point = list(A[1])

            # check if all coordinates of this point satisfy height bound
            bound_satisfied = True
            for coordinate in point:
                if coordinate.abs() > bound:
                    bound_satisfied = False
                    break
            if not bound_satisfied:
                continue

            try:
                pt = X(list(A[1]))
            except TypeError:
                pass
            else:
                rat_points.add(pt)

        return [list(_) for _ in rat_points]

    def lift_all_points():
        r"""
        Return list of all rational points lifted parallelly.
        """
        normalized_input = []
        points = modulo_points.pop() # remove the list of points corresponding to largest prime
        len_modulo_points.pop()

        for point in points:
            normalized_input.append(( (point, ), {}))
        p_iter = p_iter_fork(ncpus())
        points_satisfying = list(p_iter(parallel_function_combination, normalized_input))

        lifted_points = set()
        for pair in points_satisfying:
            L = pair[1]
            for point in L:
                lifted_points.add(X(tuple(point)))

        return list(lifted_points)

    # start of main algorithm

    primes_list = good_primes(B.ceil())

    modulo_points = points_modulo_primes(X, primes_list)
    len_modulo_points = [len(pt) for pt in modulo_points]
    len_primes = len(primes_list)
    prod_primes = prod(primes_list)

    # stores final result

    for i in range(N + 1):
        w = [0 for _ in range(N + 1)]
        w[i] = prod_primes
        m.extend(w)

    rat_points = lift_all_points()

    return sorted(rat_points)



# sage_setup: distribution = sagemath-categories
"""
Miscellaneous generic functions

A collection of functions implementing generic algorithms in arbitrary
groups, including additive and multiplicative groups.

In all cases the group operation is specified by a parameter
``operation``, which is a string either one of the set of
``multiplication_names`` or ``addition_names`` specified below, or other.
In the latter case, the caller must provide an identity, ``inverse()`` and
``op()`` functions.

::

    multiplication_names = ('multiplication', 'times', 'product', '*')
    addition_names       = ('addition', 'plus', 'sum', '+')

Also included are a generic function for computing multiples (or
powers), and an iterator for general multiples and powers.

EXAMPLES:

Some examples in the multiplicative group of a finite field:

- Discrete logs::

    sage: # needs sage.rings.finite_rings
    sage: K = GF(3^6,'b')
    sage: b = K.gen()
    sage: a = b^210
    sage: discrete_log(a, b, K.order() - 1)
    210

- Linear relation finder::

    sage: # needs sage.rings.finite_rings
    sage: F.<a> = GF(3^6,'a')
    sage: a.multiplicative_order().factor()
    2^3 * 7 * 13
    sage: b = a^7
    sage: c = a^13
    sage: linear_relation(b,c,'*')
    (13, 7)
    sage: b^13 == c^7
    True

- Orders of elements::

    sage: # needs sage.rings.finite_rings
    sage: from sage.groups.generic import order_from_multiple, order_from_bounds
    sage: k.<a> = GF(5^5)
    sage: b = a^4
    sage: order_from_multiple(b, 5^5 - 1, operation='*')
    781
    sage: order_from_bounds(b, (5^4, 5^5), operation='*')
    781

Some examples in the group of points of an elliptic curve over a finite field:

- Discrete logs::

    sage: # needs sage.libs.gap sage.rings.finite_rings sage.schemes
    sage: F = GF(37^2,'a')
    sage: E = EllipticCurve(F,[1,1])
    sage: F.<a> = GF(37^2,'a')
    sage: E = EllipticCurve(F,[1,1])
    sage: P = E(25*a + 16 , 15*a + 7 )
    sage: P.order()
    672
    sage: Q = 39*P; Q
    (36*a + 32 : 5*a + 12 : 1)
    sage: discrete_log(Q, P, P.order(), operation='+')
    39

- Linear relation finder::

    sage: # needs sage.libs.gap sage.rings.finite_rings sage.schemes
    sage: F.<a> = GF(3^6,'a')
    sage: E = EllipticCurve([a^5 + 2*a^3 + 2*a^2 + 2*a, a^4 + a^3 + 2*a + 1])
    sage: P = E(a^5 + a^4 + a^3 + a^2 + a + 2 , 0)
    sage: Q = E(2*a^3 + 2*a^2 + 2*a , a^3 + 2*a^2 + 1)
    sage: linear_relation(P,Q,'+')
    (1, 2)
    sage: P == 2*Q
    True

- Orders of elements::

    sage: # needs sage.libs.gap sage.rings.finite_rings sage.schemes
    sage: from sage.groups.generic import order_from_multiple, order_from_bounds
    sage: k.<a> = GF(5^5)
    sage: E = EllipticCurve(k,[2,4])
    sage: P = E(3*a^4 + 3*a, 2*a + 1)
    sage: M = E.cardinality(); M
    3227
    sage: plist = M.prime_factors()
    sage: order_from_multiple(P, M, plist, operation='+')
    3227
    sage: Q = E(0,2)
    sage: order_from_multiple(Q, M, plist, operation='+')
    7
    sage: order_from_bounds(Q, Hasse_bounds(5^5), operation='+')
    7
"""

# ****************************************************************************
#       Copyright (C) 2008 William Stein <wstein@gmail.com>
#                          John Cremona  <john.cremona@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from copy import copy

from sage.arith.misc import integer_ceil, integer_floor, xlcm
from sage.arith.srange import xsrange
from sage.misc.misc_c import prod
import sage.rings.integer_ring as integer_ring
import sage.rings.integer

#
# Lists of names (as strings) which the user may use to identify one
# of the standard operations:
#
multiplication_names = ('multiplication', 'times', 'product', '*')
addition_names = ('addition', 'plus', 'sum', '+')


def multiple(a, n, operation='*', identity=None, inverse=None, op=None):
    r"""
    Return either `na` or `a^n`, where `n` is any integer and `a` is
    a Python object on which a group operation such as addition or
    multiplication is defined.  Uses the standard binary algorithm.

    INPUT: See the documentation for ``discrete_logarithm()``.

    EXAMPLES::

        sage: multiple(2, 5)
        32
        sage: multiple(RealField()('2.5'), 4)                                           # needs sage.rings.real_mpfr
        39.0625000000000
        sage: multiple(2, -3)
        1/8
        sage: multiple(2, 100, '+') == 100*2
        True
        sage: multiple(2, 100) == 2**100
        True
        sage: multiple(2, -100,) == 2**-100
        True
        sage: R.<x> = ZZ[]
        sage: multiple(x, 100)
        x^100
        sage: multiple(x, 100, '+')
        100*x
        sage: multiple(x, -10)
        1/x^10

    Idempotence is detected, making the following fast::

        sage: multiple(1, 10^1000)
        1

        sage: # needs sage.schemes
        sage: E = EllipticCurve('389a1')
        sage: P = E(-1,1)
        sage: multiple(P, 10, '+')
        (645656132358737542773209599489/22817025904944891235367494656 :
         525532176124281192881231818644174845702936831/3446581505217248068297884384990762467229696 : 1)
        sage: multiple(P, -10, '+')
        (645656132358737542773209599489/22817025904944891235367494656 :
         -528978757629498440949529703029165608170166527/3446581505217248068297884384990762467229696 : 1)
    """
    from operator import inv, mul, neg, add

    if operation in multiplication_names:
        identity = a.parent().one()
        inverse = inv
        op = mul
    elif operation in addition_names:
        identity = a.parent().zero()
        inverse = neg
        op = add
    else:
        if identity is None or inverse is None or op is None:
            raise ValueError("identity, inverse and operation must all be specified")

    if n == 0:
        return identity

    if n < 0:
        n = -n
        a = inverse(a)

    if n == 1:
        return a

    # check for idempotence, and store the result otherwise
    aa = op(a, a)
    if aa == a:
        return a

    if n == 2:
        return aa

    if n == 3:
        return op(aa, a)

    if n == 4:
        return op(aa, aa)

    # since we've computed a^2, let's start squaring there
    # so, let's keep the least-significant bit around, just
    # in case.
    m = n & 1
    n = n >> 1

    # One multiplication can be saved by starting with
    # the second-smallest power needed rather than with 1
    # we've already squared a, so let's start there.
    apow = aa
    while n & 1 == 0:
        apow = op(apow, apow)
        n = n >> 1
    power = apow
    n = n >> 1

    # now multiply that least-significant bit in...
    if m:
        power = op(power, a)

    # and this is straight from the book.
    while n != 0:
        apow = op(apow, apow)
        if n & 1 != 0:
            power = op(power, apow)
        n = n >> 1

    return power


#
# Generic iterator for looping through multiples or powers
#

class multiples:
    r"""
    Return an iterator which runs through ``P0+i*P`` for ``i`` in ``range(n)``.

    ``P`` and ``P0`` must be Sage objects in some group; if the operation is
    multiplication then the returned values are instead ``P0*P**i``.

    EXAMPLES::

        sage: list(multiples(1, 10))
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        sage: list(multiples(1, 10, 100))
        [100, 101, 102, 103, 104, 105, 106, 107, 108, 109]

        sage: E = EllipticCurve('389a1')                                                # needs sage.schemes
        sage: P = E(-1,1)                                                               # needs sage.schemes
        sage: for Q in multiples(P, 5): print((Q, Q.height()/P.height()))               # needs sage.schemes
        ((0 : 1 : 0), 0.000000000000000)
        ((-1 : 1 : 1), 1.00000000000000)
        ((10/9 : -35/27 : 1), 4.00000000000000)
        ((26/361 : -5720/6859 : 1), 9.00000000000000)
        ((47503/16641 : 9862190/2146689 : 1), 16.0000000000000)

        sage: R.<x> = ZZ[]
        sage: list(multiples(x, 5))
        [0, x, 2*x, 3*x, 4*x]
        sage: list(multiples(x, 5, operation='*'))
        [1, x, x^2, x^3, x^4]
        sage: list(multiples(x, 5, indexed=True))
        [(0, 0), (1, x), (2, 2*x), (3, 3*x), (4, 4*x)]
        sage: list(multiples(x, 5, indexed=True, operation='*'))
        [(0, 1), (1, x), (2, x^2), (3, x^3), (4, x^4)]
        sage: for i,y in multiples(x, 5, indexed=True): print("%s  times %s = %s"%(i,x,y))
        0  times x = 0
        1  times x = x
        2  times x = 2*x
        3  times x = 3*x
        4  times x = 4*x

        sage: for i,n in multiples(3, 5, indexed=True, operation='*'):
        ....:     print("3 to the power %s = %s" % (i,n))
        3 to the power 0 = 1
        3 to the power 1 = 3
        3 to the power 2 = 9
        3 to the power 3 = 27
        3 to the power 4 = 81
    """
    def __init__(self, P, n, P0=None, indexed=False, operation='+', op=None):
        """
        Create a multiples iterator.

        INPUT:

        - ``P`` -- step value; any Sage object on which a binary operation is defined
        - ``n`` -- number of multiples; nonnegative integer
        - ``P0`` -- offset (default: 0); Sage object which can be 'added' to P
        - ``indexed`` -- boolean (default: ``False``)

          If ``indexed==False``, then the iterator delivers ``P0+i*P``
          (if ``operation=='+'``) or ``P0*P**i`` (if
          ``operation=='*'``), for ``i`` in ``range(n)``.

          If ``indexed==True`` then the iterator delivers tuples
          ``(i, P0+i*P)`` or ``(i, P0*P**i)``.

        - ``operation`` -- string: ``'+'`` (the default) or ``'*'`` or other.

          If other, a function ``op()`` must be supplied (a function
          of 2 arguments) defining the group binary operation; also
          ``P0`` must be supplied.
        """
        if n < 0:
            raise ValueError('n cannot be negative in multiples')

        from operator import mul, add

        if operation in multiplication_names:
            if P0 is None:
                P0 = P.parent().one()
            self.op = mul
        elif operation in addition_names:
            if P0 is None:
                P0 = P.parent().zero()
            self.op = add
        else:
            if P0 is None:
                raise ValueError("P0 must be supplied when operation is neither addition nor multiplication")
            if op is None:
                raise ValueError("op() must both be supplied when operation is neither addition nor multiplication")
            self.op = op

        self.P = copy(P)
        self.Q = copy(P0)
        if self.P is None or self.Q is None:
            raise ValueError("P and Q must not be None")
        self.i = 0
        self.bound = n
        self.indexed = indexed

    def __next__(self):
        """
        Return the next item in this multiples iterator.
        """
        if self.i >= self.bound:
            raise StopIteration
        i = self.i
        val = self.Q
        self.i += 1
        self.Q = self.op(self.Q, self.P)
        if self.indexed:
            return (i, val)
        else:
            return val

    next = __next__

    def __iter__(self):
        """
        Standard member function making this class an iterator.
        """
        return self


def bsgs(a, b, bounds, operation='*', identity=None, inverse=None, op=None):
    r"""
    Totally generic discrete baby-step giant-step function.

    Solves `na=b` (or `a^n=b`) with `lb\le n\le ub` where ``bounds==(lb,ub)``,
    raising an error if no such `n` exists.

    `a` and `b` must be elements of some group with given identity,
    inverse of ``x`` given by ``inverse(x)``, and group operation on
    ``x``, ``y`` by ``op(x,y)``.

    If operation is '*' or '+' then the other
    arguments are provided automatically; otherwise they must be
    provided by the caller.

    INPUT:

    - ``a`` -- group element
    - ``b`` -- group element
    - ``bounds`` -- a 2-tuple of integers ``(lower,upper)`` with ``0<=lower<=upper``
    - ``operation`` -- string: ``'*'``, ``'+'``, other
    - ``identity`` -- the identity element of the group
    - ``inverse`` -- function of 1 argument ``x``, returning inverse of ``x``
    - ``op`` -- function of 2 arguments ``x``, ``y`` returning ``x*y`` in the group

    OUTPUT:

    An integer `n` such that `a^n = b` (or `na = b`).  If no
    such `n` exists, this function raises a :exc:`ValueError` exception.

    .. NOTE::

        This is a generalization of discrete logarithm.  One
        situation where this version is useful is to find the order of
        an element in a group where we only have bounds on the group
        order (see the elliptic curve example below).

    ALGORITHM: Baby step giant step.  Time and space are soft
    `O(\sqrt{n})` where `n` is the difference between upper and lower
    bounds.

    EXAMPLES::

        sage: from sage.groups.generic import bsgs
        sage: b = Mod(2,37);  a = b^20
        sage: bsgs(b, a, (0, 36))
        20

        sage: p = next_prime(10^20)                                                     # needs sage.libs.pari
        sage: a = Mod(2,p); b = a^(10^25)                                               # needs sage.libs.pari
        sage: bsgs(a, b, (10^25 - 10^6, 10^25 + 10^6)) == 10^25                         # needs sage.libs.pari
        True

        sage: # needs sage.rings.finite_rings
        sage: K = GF(3^6,'b')
        sage: a = K.gen()
        sage: b = a^210
        sage: bsgs(a, b, (0, K.order() - 1))
        210

        sage: K.<z> = CyclotomicField(230)                                              # needs sage.rings.number_field
        sage: w = z^500                                                                 # needs sage.rings.number_field
        sage: bsgs(z, w, (0, 229))                                                      # needs sage.rings.number_field
        40

    An additive example in an elliptic curve group::

        sage: # needs sage.rings.finite_rings sage.schemes
        sage: F.<a> = GF(37^5)
        sage: E = EllipticCurve(F, [1,1])
        sage: P = E.lift_x(a); P
        (a : 28*a^4 + 15*a^3 + 14*a^2 + 7 : 1)

    This will return a multiple of the order of P::

        sage: bsgs(P, P.parent().zero(), Hasse_bounds(F.order()), operation='+')            # needs sage.rings.finite_rings sage.schemes
        69327408

    AUTHOR:

    - John Cremona (2008-03-15)
    """
    Z = integer_ring.ZZ

    from operator import inv, mul, neg, add

    if operation in multiplication_names:
        identity = a.parent().one()
        inverse = inv
        op = mul
    elif operation in addition_names:
        # Should this be replaced with .zero()? With an extra AttributeError handler?
        identity = a.parent().zero()
        inverse = neg
        op = add
    else:
        if identity is None or inverse is None or op is None:
            raise ValueError("identity, inverse and operation must be given")

    lb, ub = bounds
    if lb < 0 or ub < lb:
        raise ValueError("bsgs() requires 0<=lb<=ub")

    if a == identity and b != identity:
        raise ValueError("no solution in bsgs()")

    ran = 1 + ub - lb   # the length of the interval

    mult = lambda x, y: multiple(x, y, operation=operation, identity=identity, inverse=inverse, op=op)
    c = op(inverse(b), mult(a, lb))

    if ran < 30:    # use simple search for small ranges
        d = c
        # for i,d in multiples(a,ran,c,indexed=True,operation=operation):
        for i0 in range(ran):
            i = lb + i0
            if identity == d:        # identity == b^(-1)*a^i, so return i
                return Z(i)
            d = op(a, d)
        raise ValueError("no solution in bsgs()")

    m = ran.isqrt() + 1  # we need sqrt(ran) rounded up
    table = {}       # will hold pairs (a^(lb+i),lb+i) for i in range(m)

    d = c
    for i0 in xsrange(m):
        i = lb + i0
        if identity == d:        # identity == b^(-1)*a^i, so return i
            return Z(i)
        table[d] = i
        d = op(d, a)

    c = op(c, inverse(d))     # this is now a**(-m)
    d = identity
    for i in xsrange(m):
        j = table.get(d)
        if j is not None:  # then d == b*a**(-i*m) == a**j
            return Z(i * m + j)
        d = op(c, d)

    raise ValueError("log of %s to the base %s does not exist in %s" % (b, a, bounds))


def discrete_log_rho(a, base, ord=None, operation='*', identity=None, inverse=None, op=None, hash_function=hash):
    """
    Pollard Rho algorithm for computing discrete logarithm in cyclic
    group of prime order.
    If the group order is very small it falls back to the baby step giant step
    algorithm.

    INPUT:

    - ``a`` -- a group element
    - ``base`` -- a group element
    - ``ord`` -- the order of ``base`` or ``None``, in this case we try
      to compute it
    - ``operation`` -- string (default: ``'*'``); denoting whether we
      are in an additive group or a multiplicative one
    - ``identity`` -- the group's identity
    - ``inverse`` -- function of 1 argument ``x``, returning inverse of ``x``
    - ``op`` -- function of 2 arguments ``x``, ``y``, returning ``x*y`` in the group
    - ``hash_function`` -- having an efficient hash function is critical
      for this algorithm (see examples)

    OUTPUT: integer `n` such that `a = base^n` (or `a = n*base`)

    ALGORITHM: Pollard rho for discrete logarithm, adapted from the
    article of Edlyn Teske, 'A space efficient algorithm for group
    structure computation'.

    EXAMPLES::

        sage: F.<a> = GF(2^13)                                                          # needs sage.rings.finite_rings
        sage: g = F.gen()                                                               # needs sage.rings.finite_rings
        sage: discrete_log_rho(g^1234, g)                                               # needs sage.rings.finite_rings
        1234

        sage: # needs sage.rings.finite_rings sage.schemes
        sage: F.<a> = GF(37^5)
        sage: E = EllipticCurve(F, [1,1])
        sage: G = (3*31*2^4)*E.lift_x(a)
        sage: discrete_log_rho(12345*G, G, ord=46591, operation='+')
        12345

    It also works with matrices::

        sage: A = matrix(GF(50021), [[10577, 23999, 28893],                             # needs sage.modules sage.rings.finite_rings
        ....:                        [14601, 41019, 30188],
        ....:                        [3081, 736, 27092]])
        sage: discrete_log_rho(A^1234567, A)                                            # needs sage.modules sage.rings.finite_rings
        1234567

    Beware, the order must be prime::

        sage: I = IntegerModRing(171980)
        sage: discrete_log_rho(I(2), I(3))                                              # needs sage.libs.pari
        Traceback (most recent call last):
        ...
        ValueError: for Pollard rho algorithm the order of the group must be prime

    If it fails to find a suitable logarithm, it raises a :exc:`ValueError`::

        sage: I = IntegerModRing(171980)
        sage: discrete_log_rho(I(31002), I(15501))                                      # needs sage.libs.pari
        Traceback (most recent call last):
        ...
        ValueError: Pollard rho algorithm failed to find a logarithm

    The main limitation on the hash function is that we don't want to have
    ``hash(x*y) == hash(x) + hash(y)``::

        sage: # needs sage.libs.pari
        sage: I = IntegerModRing(next_prime(2^23))
        sage: def test():
        ....:     try:
        ....:          discrete_log_rho(I(123456), I(1), operation='+')
        ....:     except Exception:
        ....:          print("FAILURE")
        sage: test()  # random failure
        FAILURE

    If this happens, we can provide a better hash function::

        sage: discrete_log_rho(I(123456), I(1), operation='+',                          # needs sage.libs.pari
        ....:                  hash_function=lambda x: hash(x*x))
        123456

    AUTHOR:

    - Yann Laigle-Chapuy (2009-09-05)
    """
    from sage.rings.integer import Integer
    from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
    from operator import mul, add, pow

    # should be reasonable choices
    partition_size = 20
    memory_size = 4
    mult = op
    power = lambda x, y: multiple(x, y, operation=operation, identity=identity, inverse=inverse, op=op)
    if operation in addition_names:
        mult = add
        power = mul
        if ord is None:
            ord = base.additive_order()
    elif operation in multiplication_names:
        mult = mul
        power = pow
        if ord is None:
            ord = base.multiplicative_order()
    elif ord is None or inverse is None or identity is None or op is None:
        raise ValueError

    ord = Integer(ord)
    if not ord.is_prime():
        raise ValueError("for Pollard rho algorithm the order of the group must be prime")

    # check if we need to set immutable before hashing
    mut = hasattr(base, 'set_immutable')

    isqrtord = ord.isqrt()

    if isqrtord < partition_size:  # setup to costly, use bsgs
        return bsgs(base, a, bounds=(0, ord), identity=identity, inverse=inverse, op=op, operation=operation)

    reset_bound = 8 * isqrtord  # we take some margin

    I = IntegerModRing(ord)

    for s in range(10):  # to avoid infinite loops
        # random walk function setup
        m = [I.random_element() for i in range(partition_size)]
        n = [I.random_element() for i in range(partition_size)]
        M = [mult(power(base, Integer(m[i])), power(a, Integer(n[i])))
             for i in range(partition_size)]

        ax = I.random_element()
        x = power(base, Integer(ax))
        if mut:
            x.set_immutable()

        bx = I(0)

        sigma = [(0, None)] * memory_size
        H = {}  # memory
        i0 = 0
        nextsigma = 0
        for i in range(reset_bound):
            # random walk, we need an efficient hash
            s = hash_function(x) % partition_size
            x, ax, bx = (mult(M[s], x), ax + m[s], bx + n[s])
            if mut:
                x.set_immutable()
            # look for collisions
            if x in H:
                ay, by = H[x]
                if bx == by:
                    break
                else:
                    res = sage.rings.integer.Integer((ay - ax) / (bx - by))
                    if power(base, res) == a:
                        return res
                    else:
                        break
            # should we remember this value?
            elif i >= nextsigma:
                if sigma[i0][1] is not None:
                    H.pop(sigma[i0][1])
                sigma[i0] = (i, x)
                i0 = (i0 + 1) % memory_size
                nextsigma = 3 * sigma[i0][0]  # 3 seems a good choice
                H[x] = (ax, bx)

    raise ValueError("Pollard rho algorithm failed to find a logarithm")


def discrete_log(a, base, ord=None, bounds=None, operation='*', identity=None, inverse=None, op=None, algorithm='bsgs'):
    r"""
    Totally generic discrete log function.

    INPUT:

    - ``a`` -- group element
    - ``base`` -- group element (the base)
    - ``ord`` -- integer (multiple of order of base, or ``None``)
    - ``bounds`` -- a priori bounds on the log
    - ``operation`` -- string: ``'*'``, ``'+'``, other
    - ``identity`` -- the group's identity
    - ``inverse`` -- function of 1 argument ``x``, returning inverse of ``x``
    - ``op`` -- function of 2 arguments ``x``, ``y``, returning ``x*y`` in the group
    - ``algorithm`` -- string denoting what algorithm to use for prime-order
      logarithms: ``'bsgs'``, ``'rho'``, ``'lambda'``

    ``a`` and ``base`` must be elements of some group with identity
    given by ``identity``, inverse of ``x`` by ``inverse(x)``, and group
    operation on ``x``, ``y`` by ``op(x,y)``.

    If operation is ``'*'`` or ``'+'``, then the other
    arguments are provided automatically; otherwise they must be
    provided by the caller.

    OUTPUT:

    This returns an integer `n` such that `b^n = a` (or `nb = a`),
    assuming that ``ord`` is a multiple of the order of the base `b`.
    If ``ord`` is not specified, an attempt is made to compute it.

    If no such `n` exists, this function raises a :exc:`ValueError` exception.

    .. WARNING::

       If ``x`` has a ``log`` method, it is likely to be vastly faster
       than using this function.  E.g., if ``x`` is an integer modulo
       `n`, use its ``log`` method instead!

    ALGORITHM: Pohlig-Hellman, Baby step giant step, Pollard's lambda/kangaroo, and Pollard's rho.

    EXAMPLES::

        sage: b = Mod(2,37);  a = b^20
        sage: discrete_log(a, b)
        20
        sage: b = Mod(3,2017);  a = b^20
        sage: discrete_log(a, b, bounds=(10, 100))
        20

        sage: # needs sage.rings.finite_rings
        sage: K = GF(3^6, 'b')
        sage: b = K.gen()
        sage: a = b^210
        sage: discrete_log(a, b, K.order() - 1)
        210

        sage: b = Mod(1,37);  x = Mod(2,37)
        sage: discrete_log(x, b)
        Traceback (most recent call last):
        ...
        ValueError: no discrete log of 2 found to base 1
        sage: b = Mod(1,997);  x = Mod(2,997)
        sage: discrete_log(x, b)
        Traceback (most recent call last):
        ...
        ValueError: no discrete log of 2 found to base 1

    See :issue:`2356`::

        sage: F.<w> = GF(121)                                                           # needs sage.rings.finite_rings
        sage: v = w^120                                                                 # needs sage.rings.finite_rings
        sage: v.log(w)                                                                  # needs sage.rings.finite_rings
        0

        sage: K.<z> = CyclotomicField(230)                                              # needs sage.rings.number_field
        sage: w = z^50                                                                  # needs sage.rings.number_field
        sage: discrete_log(w, z)                                                        # needs sage.rings.number_field
        50

    An example where the order is infinite: note that we must give
    an upper bound here::

        sage: # needs sage.rings.number_field
        sage: K.<a> = QuadraticField(23)
        sage: eps = 5*a - 24        # a fundamental unit
        sage: eps.multiplicative_order()
        +Infinity
        sage: eta = eps^100
        sage: discrete_log(eta, eps, bounds=(0,1000))
        100

    In this case we cannot detect negative powers::

        sage: eta = eps^(-3)                                                            # needs sage.rings.number_field
        sage: discrete_log(eta,eps,bounds=(0,100))                                      # needs sage.rings.number_field
        Traceback (most recent call last):
        ...
        ValueError: no discrete log of -11515*a - 55224 found to base 5*a - 24

    But we can invert the base (and negate the result) instead::

        sage: -discrete_log(eta^-1, eps, bounds=(0,100))                                # needs sage.rings.number_field
        -3

    An additive example: elliptic curve DLOG::

        sage: # needs sage.libs.gap sage.rings.finite_rings sage.schemes
        sage: F = GF(37^2,'a')
        sage: E = EllipticCurve(F, [1,1])
        sage: F.<a> = GF(37^2,'a')
        sage: E = EllipticCurve(F, [1,1])
        sage: P = E(25*a + 16, 15*a + 7)
        sage: P.order()
        672
        sage: Q = 39*P; Q
        (36*a + 32 : 5*a + 12 : 1)
        sage: discrete_log(Q, P, P.order(), operation='+')
        39

    An example of big smooth group::

        sage: # needs sage.rings.finite_rings
        sage: F.<a> = GF(2^63)
        sage: g = F.gen()
        sage: u = g**123456789
        sage: discrete_log(u,g)
        123456789

    The above examples also work when the ``'rho'`` and ``'lambda'`` algorithms are used::

        sage: b = Mod(2,37);  a = b^20
        sage: discrete_log(a, b, algorithm='rho')
        20
        sage: b = Mod(3,2017);  a = b^20
        sage: discrete_log(a, b, algorithm='lambda', bounds=(10, 100))
        20

        sage: # needs sage.rings.finite_rings
        sage: K = GF(3^6,'b')
        sage: b = K.gen()
        sage: a = b^210
        sage: discrete_log(a, b, K.order()-1, algorithm='rho')
        210

        sage: b = Mod(1,37);  x = Mod(2,37)
        sage: discrete_log(x, b, algorithm='lambda')
        Traceback (most recent call last):
        ...
        ValueError: no discrete log of 2 found to base 1
        sage: b = Mod(1,997);  x = Mod(2,997)
        sage: discrete_log(x, b, algorithm='rho')
        Traceback (most recent call last):
        ...
        ValueError: no discrete log of 2 found to base 1

        sage: # needs sage.libs.gap sage.rings.finite_rings sage.schemes
        sage: F = GF(37^2,'a')
        sage: E = EllipticCurve(F, [1,1])
        sage: F.<a> = GF(37^2,'a')
        sage: E = EllipticCurve(F, [1,1])
        sage: P = E(25*a + 16, 15*a + 7)
        sage: P.order()
        672
        sage: Q = 39*P; Q
        (36*a + 32 : 5*a + 12 : 1)
        sage: discrete_log(Q, P, P.order(), operation='+', algorithm='lambda')
        39

        sage: # needs sage.rings.finite_rings
        sage: F.<a> = GF(2^63)
        sage: g = F.gen()
        sage: u = g**123456789
        sage: discrete_log(u, g, algorithm='rho')
        123456789

    TESTS:

    Random testing::

        sage: G = Zmod(randrange(1, 1000))
        sage: base = G.random_element()
        sage: order = choice([base.additive_order(), G.order()])
        sage: assert order.divides(G.cardinality())
        sage: sol = randrange(base.additive_order())
        sage: elem = sol * base
        sage: args = (elem, base, order)
        sage: kwargs = {'operation': '+'}
        sage: kwargs['algorithm'] = choice(['bsgs', 'rho', 'lambda'])
        sage: if randrange(2):
        ....:     lo = randrange(-order, sol + 1)
        ....:     hi = randrange(sol + 1, 2*order)
        ....:     assert lo <= sol <= hi
        ....:     kwargs['bounds'] = (lo, hi)
        sage: try:
        ....:     res = discrete_log(*args, **kwargs)
        ....: except ValueError:
        ....:     # lambda can fail randomly
        ....:     assert kwargs['algorithm'] == 'lambda'
        ....: else:
        ....:     assert res == sol

    AUTHORS:

    - William Stein and David Joyner (2005-01-05)
    - John Cremona (2008-02-29) rewrite using ``dict()`` and make generic
    - Julien Grijalva (2022-08-09) rewrite to make more generic, more algorithm options, and more effective use of bounds
    """
    from operator import mul, add, pow
    power = mul if operation in addition_names else pow
    mult = add if operation in addition_names else mul
    if op:
        mult = op
        power = lambda x, y: multiple(x, y, operation=operation, identity=identity, inverse=inverse, op=op)
    if bounds:
        lb, ub = map(integer_ring.ZZ, bounds)
    if (op is None or identity is None or inverse is None or ord is None) and operation not in addition_names + multiplication_names:
        raise ValueError("ord, op, identity, and inverse must all be specified for this operation")
    if ord is None:
        if operation in multiplication_names:
            try:
                ord = base.multiplicative_order()
            except Exception:
                ord = base.order()
        else:
            try:
                ord = base.additive_order()
            except Exception:
                ord = base.order()
    else:
        ord = integer_ring.ZZ(ord)
    try:
        from sage.rings.infinity import Infinity
        if ord == +Infinity:
            return bsgs(base, a, bounds, identity=identity, inverse=inverse, op=op, operation=operation)
        if base == power(base, 0) and a != base:
            raise ValueError
        f = ord.factor()
        l = [0] * len(f)
        mods = []
        running_mod = 1
        offset = 0
        if bounds:
            a = mult(a, power(base, -lb))
            offset = lb
            bound = ub - lb
        i = -1  # this corrects a bug in which the loop is never entered and i never gets assigned a value
        for i, (pi, ri) in enumerate(f):
            gamma = power(base, ord // pi)
            # pohlig-hellman doesn't work with an incorrect order, and the user might have provided a bad parameter
            while gamma == power(gamma, 0) and ri > 0:  # identity might be None
                ord //= pi
                ri -= 1
                gamma = power(base, ord // pi)
            if not bounds:
                bound = ord - 1
            running_bound = min(bound, pi**ri - 1)
            j = -1
            for j in range(ri):
                temp_bound = min(running_bound, pi - 1)
                h = power(mult(a, power(base, -l[i])), ord // pi**(j + 1))
                if algorithm == 'bsgs':
                    c = bsgs(gamma, h, (0, temp_bound), inverse=inverse, identity=identity, op=op, operation=operation)
                elif algorithm == 'rho':
                    c = discrete_log_rho(h, gamma, ord=pi, inverse=inverse, identity=identity, op=op, operation=operation)
                elif algorithm == 'lambda':
                    c = discrete_log_lambda(h, gamma, (0, temp_bound), inverse=inverse, identity=identity, op=op, operation=operation)
                l[i] += c * (pi**j)
                running_bound //= pi
                running_mod *= pi
                if running_mod > bound:
                    break
            mods.append(pi ** (j+1))
            if running_mod > bound:
                break  # we have log%running_mod. if we know that log<running_mod, then we have the value of log.
        l = l[:i + 1]
        from sage.arith.misc import CRT_list
        return (CRT_list(l, mods) + offset) % ord
    except ValueError:
        raise ValueError("no discrete log of %s found to base %s" % (a, base))


def discrete_log_generic(a, base, ord=None, bounds=None, operation='*', identity=None, inverse=None, op=None, algorithm='bsgs'):
    """
    Alias for ``discrete_log``.
    """
    return discrete_log(a, base, ord=ord, bounds=bounds, operation=operation, identity=identity, inverse=inverse, op=op, algorithm=algorithm)


def discrete_log_lambda(a, base, bounds, operation='*', identity=None, inverse=None, op=None, hash_function=hash):
    """
    Pollard Lambda algorithm for computing discrete logarithms. It uses
    only a logarithmic amount of memory. It's useful if you have
    bounds on the logarithm. If you are computing logarithms in a
    whole finite group, you should use Pollard Rho algorithm.

    INPUT:

    - ``a`` -- a group element
    - ``base`` -- a group element
    - ``bounds`` -- a couple (lb,ub) representing the range where we look for a logarithm
    - ``operation`` -- string: '+', '*' or 'other'
    - ``identity`` -- the identity element of the group
    - ``inverse`` -- function of 1 argument ``x`` returning inverse of ``x``
    - ``op`` -- function of 2 arguments ``x``, ``y`` returning ``x*y`` in the group
    - ``hash_function`` -- having an efficient hash function is critical for this algorithm

    OUTPUT: integer `n` such that `a=base^n` (or `a=n*base`)

    ALGORITHM: Pollard Lambda, if bounds are (lb,ub) it has time complexity
        O(sqrt(ub-lb)) and space complexity O(log(ub-lb))

    EXAMPLES::

        sage: F.<a> = GF(2^63)                                                          # needs sage.rings.finite_rings
        sage: discrete_log_lambda(a^1234567, a, (1200000,1250000))                      # needs sage.rings.finite_rings
        1234567

        sage: # needs sage.rings.finite_rings sage.schemes
        sage: F.<a> = GF(37^5)
        sage: E = EllipticCurve(F, [1,1])
        sage: P = E.lift_x(a); P
        (a : 28*a^4 + 15*a^3 + 14*a^2 + 7 : 1)

    This will return a multiple of the order of P::

        sage: discrete_log_lambda(P.parent().zero(), P, Hasse_bounds(F.order()),            # needs sage.rings.finite_rings sage.schemes
        ....:                     operation='+')
        69327408

        sage: K.<a> = GF(89**5)                                                         # needs sage.rings.finite_rings
        sage: hs = lambda x: hash(x) + 15
        sage: discrete_log_lambda(a**(89**3 - 3),       # long time (10s on sage.math, 2011), needs sage.rings.finite_rings
        ....:                     a, (89**2, 89**4), operation='*', hash_function=hs)
        704966

    AUTHOR:

        -- Yann Laigle-Chapuy (2009-01-25)
    """
    from sage.rings.integer import Integer
    from operator import mul, add

    mult = op
    if operation in addition_names:
        mult = add
    elif operation in multiplication_names:
        mult = mul
    power = lambda x, y: multiple(x, y, operation=operation, identity=identity, inverse=inverse, op=op)

    lb, ub = bounds
    if lb < 0 or ub < lb:
        raise ValueError("discrete_log_lambda() requires 0<=lb<=ub")

    # check for mutability
    mut = hasattr(base, 'set_immutable')

    width = Integer(ub - lb)
    N = width.isqrt() + 1

    M = {}
    for s in range(10):  # to avoid infinite loops
        # random walk function setup
        k = 0
        while 2**k < N:
            r = sage.misc.prandom.randrange(1, N)
            M[k] = (r, power(base, r))
            k += 1
        # first random walk
        H = power(base, ub)
        c = ub
        for i in range(N):
            if mut:
                H.set_immutable()
            r, e = M[hash_function(H) % k]
            H = mult(H, e)
            c += r
        if mut:
            H.set_immutable()
        mem = {H}
        # second random walk
        H = a
        d = 0
        while c - d >= lb:
            if mut:
                H.set_immutable()
            if ub >= c - d and H in mem:
                return c - d
            r, e = M[hash_function(H) % k]
            H = mult(H, e)
            d += r

    raise ValueError("Pollard Lambda failed to find a log")


################################################################
#
# Generic linear relation finder
#
################################################################


def linear_relation(P, Q, operation='+', identity=None, inverse=None, op=None):
    r"""
    Function which solves the equation ``a*P=m*Q`` or ``P^a=Q^m``.

    Additive version: returns `(a,m)` with minimal `m>0` such that
    `aP=mQ`.  Special case: if `\left<P\right>` and `\left<Q\right>`
    intersect only in `\{0\}` then `(a,m)=(0,n)` where `n` is
    ``Q.additive_order()``.

    Multiplicative version: returns `(a,m)` with minimal `m>0` such
    that `P^a=Q^m`.  Special case: if `\left<P\right>` and
    `\left<Q\right>` intersect only in `\{1\}` then `(a,m)=(0,n)`
    where `n` is ``Q.multiplicative_order()``.

    ALGORITHM:

    Uses the generic ``bsgs()`` function, and so works in general
    finite abelian groups.

    EXAMPLES:

    An additive example (in an elliptic curve group)::

        sage: # needs sage.rings.finite_rings sage.schemes
        sage: F.<a> = GF(3^6,'a')
        sage: E = EllipticCurve([a^5 + 2*a^3 + 2*a^2 + 2*a, a^4 + a^3 + 2*a + 1])
        sage: P = E(a^5 + a^4 + a^3 + a^2 + a + 2, 0)
        sage: Q = E(2*a^3 + 2*a^2 + 2*a, a^3 + 2*a^2 + 1)
        sage: linear_relation(P, Q, '+')
        (1, 2)
        sage: P == 2*Q
        True

    A multiplicative example (in a finite field's multiplicative group)::

        sage: # needs sage.rings.finite_rings
        sage: F.<a> = GF(3^6,'a')
        sage: a.multiplicative_order().factor()
        2^3 * 7 * 13
        sage: b = a^7
        sage: c = a^13
        sage: linear_relation(b, c, '*')
        (13, 7)
        sage: b^13 == c^7
        True
    """
    Z = integer_ring.ZZ

    if operation in multiplication_names:
        try:
            n = P.multiplicative_order()
            m = Q.multiplicative_order()
        except Exception:
            n = P.order()
            m = Q.order()
    elif operation in addition_names:
        try:
            n = P.additive_order()
            m = Q.additive_order()
        except Exception:
            n = P.order()
            m = Q.order()
    else:
        if op is None:
            raise ValueError("operation must be specified")
        n = P.order()
        m = Q.order()

    g = n.gcd(m)
    if g == 1:
        return (m, Z.zero())
    n1 = n // g
    m1 = m // g
    P1 = multiple(P, n1, operation=operation)  # has exact order g
    Q1 = multiple(Q, m1, operation=operation)  # has exact order g

    # now see if Q1 is a multiple of P1; the only multiples we
    # need check are h*Q1 where h divides g
    for h in g.divisors():  # positive divisors!
        try:
            Q2 = multiple(Q1, h, operation=operation)
            return (n1 * bsgs(P1, Q2, (0, g - 1), operation=operation),
                    m1 * h)
        except ValueError:
            pass  # to next h
    raise ValueError("no solution found in linear_relation")

################################################################
#
# Generic functions to find orders of elements
#
# 1. order_from_multiple: finds the order given a multiple of the order
#
# 2. order_from_bounds: finds the order given an interval containing a
# multiple of the order
#
# 3. has_order: check if the order is exactly equal to a given integer
#
################################################################


def order_from_multiple(P, m, plist=None, factorization=None, check=True,
                        operation='+', identity=None, inverse=None, op=None):
    r"""
    Generic function to find order of a group element given a multiple
    of its order.

    See :meth:`bsgs` for full explanation of the inputs.

    INPUT:

    - ``P`` -- a Sage object which is a group element
    - ``m`` -- Sage integer which is a multiple of the order of ``P``,
      i.e. we require that ``m*P=0`` (or ``P**m=1``)
    - ``check`` -- a Boolean (default: ``True``), indicating whether we check if ``m``
      really is a multiple of the order
    - ``factorization`` -- the factorization of ``m``, or ``None`` in which
      case this function will need to factor ``m``
    - ``plist`` -- list of the prime factors of ``m``, or ``None``. Kept for compatibility only,
      prefer the use of ``factorization``
    - ``operation`` -- string: ``'+'`` (default), ``'*'`` or ``None``
    - ``identity`` -- the identity element of the group
    - ``inverse`` -- function of 1 argument ``x``, returning inverse of ``x``
    - ``op`` -- function of 2 arguments ``x``, ``y`` returning ``x*y`` in the group

    .. NOTE::

        It is more efficient for the caller to factor ``m`` and cache
        the factors for subsequent calls.

    EXAMPLES::

        sage: from sage.groups.generic import order_from_multiple

        sage: # needs sage.rings.finite_rings
        sage: k.<a> = GF(5^5)
        sage: b = a^4
        sage: order_from_multiple(b, 5^5 - 1, operation='*')
        781

        sage: # needs sage.rings.finite_rings sage.schemes
        sage: E = EllipticCurve(k, [2,4])
        sage: P = E(3*a^4 + 3*a, 2*a + 1)
        sage: M = E.cardinality(); M
        3227
        sage: F = M.factor()
        sage: order_from_multiple(P, M, factorization=F, operation='+')
        3227
        sage: Q = E(0,2)
        sage: order_from_multiple(Q, M, factorization=F, operation='+')
        7

        sage: # needs sage.rings.number_field
        sage: K.<z> = CyclotomicField(230)
        sage: w = z^50
        sage: order_from_multiple(w, 230, operation='*')
        23

        sage: # needs sage.modules sage.rings.finite_rings
        sage: F = GF(2^1279,'a')
        sage: n = F.cardinality() - 1  # Mersenne prime
        sage: order_from_multiple(F.random_element(), n,
        ....:                     factorization=[(n,1)], operation='*') == n
        True

        sage: # needs sage.rings.finite_rings
        sage: K.<a> = GF(3^60)
        sage: order_from_multiple(a, 3^60 - 1, operation='*', check=False)
        42391158275216203514294433200

    TESTS:

    Check that :issue:`38489` is fixed::

        sage: from sage.groups.generic import order_from_multiple
        sage: plist = [43, 257, 547, 881]
        sage: m = prod(plist[:-1])
        sage: elt = Zmod(m)(plist[-1])
        sage: order_from_multiple(elt, m, plist=plist)
        6044897
    """
    Z = integer_ring.ZZ

    if operation in multiplication_names:
        identity = P.parent().one()
    elif operation in addition_names:
        identity = P.parent().zero()
    else:
        if identity is None or inverse is None or op is None:
            raise ValueError("identity, inverse and operation must all be specified")

    def _multiple(A, B):
        return multiple(A,
                        B,
                        operation=operation,
                        identity=identity,
                        inverse=inverse,
                        op=op)

    if P == identity:
        return Z.one()

    M = Z(m)
    if check and _multiple(P, M) != identity:
        raise ValueError(f"The order of P(={P}) does not divide {M}")

    if factorization:
        F = factorization
    elif plist:
        F = [(p, M.valuation(p)) for p in plist]
    else:
        F = M.factor()

    if len(F) == 1 and list(F) == [(M, 1)]:
        return M

    # Efficiency improvement (2009-10-27, implemented by Yann Laigle-Chapuy):
    # we use an internal recursive function to avoid unnecessary computations.
    def _order_from_multiple_helper(Q, L, S):
        """
        For internal use, to minimize the number of group operations.
        """
        l = len(L)
        if l == 1:
            # we determine the power of p dividing the order,

            # Efficiency improvement (2009-04-01, suggested by Ryan Hinton,
            # implemented by John Cremona): avoid the last multiplication by p.
            # For example, if M itself is prime the code used to compute M*P
            # twice (unless P=0), now it does it once.
            p, e = L[0]
            e0 = 0
            while (Q != identity) and (e0 < e - 1):
                Q = _multiple(Q, p)
                e0 += 1
            if Q != identity:
                e0 += 1
            return p**e0
        else:
            # try to split the list wisely
            sum_left = 0
            for k in range(l):
                p, e = L[k]
                # multiplying by p**e require roughly 'e log_2(p) / 2' additions
                v = e * sage.functions.log.log(float(p))
                if abs(sum_left + v - (S / 2)) > abs(sum_left - (S / 2)):
                    break
                sum_left += v
            if not 0 < k < l:
                k = l // 2
            L1 = L[:k]
            L2 = L[k:]
            # recursive calls
            o1 = _order_from_multiple_helper(
                _multiple(Q, prod([p**e for p, e in L2])), L1, sum_left)
            o2 = _order_from_multiple_helper(_multiple(Q, o1), L2, S - sum_left)
            return o1 * o2

    return _order_from_multiple_helper(P, F, sage.functions.log.log(float(M)))


def order_from_bounds(P, bounds, d=None, operation='+',
                      identity=None, inverse=None, op=None):
    r"""
    Generic function to find order of a group element, given only
    upper and lower bounds for a multiple of the order (e.g. bounds on
    the order of the group of which it is an element)

    INPUT:

    - ``P`` -- a Sage object which is a group element

    - ``bounds`` -- a 2-tuple ``(lb,ub)`` such that ``m*P=0`` (or
      ``P**m=1``) for some ``m`` with ``lb<=m<=ub``

    - ``d`` -- (optional) a positive integer; only ``m`` which are
      multiples of this will be considered

    - ``operation`` -- string; ``'+'`` (default) or ``'*'`` or other.
      If other, the following must be supplied:

      - ``identity`` -- the identity element for the group
      - ``inverse`` -- a function of one argument giving the inverse
        of a group element
      - ``op`` -- a function of 2 arguments defining the group binary
        operation

    .. NOTE::

       Typically ``lb`` and ``ub`` will be bounds on the group order,
       and from previous calculation we know that the group order is
       divisible by ``d``.

    EXAMPLES::

        sage: from sage.groups.generic import order_from_bounds

        sage: # needs sage.rings.finite_rings
        sage: k.<a> = GF(5^5)
        sage: b = a^4
        sage: order_from_bounds(b, (5^4, 5^5), operation='*')
        781

        sage: # needs sage.rings.finite_rings sage.schemes
        sage: E = EllipticCurve(k, [2,4])
        sage: P = E(3*a^4 + 3*a, 2*a + 1)
        sage: bounds = Hasse_bounds(5^5)
        sage: Q = E(0,2)
        sage: order_from_bounds(Q, bounds, operation='+')
        7
        sage: order_from_bounds(P, bounds, 7, operation='+')
        3227

        sage: # needs sage.rings.number_field
        sage: K.<z> = CyclotomicField(230)
        sage: w = z^50
        sage: order_from_bounds(w, (200, 250), operation='*')
        23
    """
    from operator import mul, add

    if operation in multiplication_names:
        op = mul
        identity = P.parent().one()
    elif operation in addition_names:
        op = add
        identity = P.parent().zero()
    else:
        if op is None:
            raise ValueError("operation and identity must be specified")

    Q = P
    if d is None:
        d = 1
    if d > 1:
        Q = multiple(P, d, operation=operation)
        lb, ub = bounds
        bounds = (integer_ceil(lb / d), integer_floor(ub / d))

    # Use generic bsgs to find  n=d*m with lb<=n<=ub and n*P=0

    m = d * bsgs(Q, identity, bounds, operation=operation)

    # Now use the order_from_multiple() function to finish the job:

    return order_from_multiple(P, m, operation=operation, check=False)


def has_order(P, n, operation='+'):
    r"""
    Generic function to test if a group element `P` has order
    exactly equal to a given positive integer `n`.

    INPUT:

    - ``P`` -- group element with respect to the specified ``operation``
    - ``n`` -- positive integer, or its :class:`~sage.structure.factorization.Factorization`
    - ``operation`` -- string, either ``'+'`` (default) or ``'*'``

    EXAMPLES::

        sage: from sage.groups.generic import has_order

        sage: # needs sage.schemes
        sage: E.<P> = EllipticCurve(GF(71), [5,5])
        sage: P.order()
        57
        sage: has_order(P, 57)
        True
        sage: has_order(P, factor(57))
        True
        sage: has_order(P, 19)
        False
        sage: has_order(3*P, 19)
        True
        sage: has_order(3*P, 57)
        False

    ::

        sage: R = Zmod(14981)
        sage: g = R(321)
        sage: g.multiplicative_order()                                                  # needs sage.libs.pari
        42
        sage: has_order(g, 42, operation='*')
        True
        sage: has_order(g, factor(42), operation='*')
        True
        sage: has_order(g, 70, operation='*')
        False

    TESTS::

        sage: # needs sage.libs.pari sage.modules
        sage: ns = [randrange(1,10**5) for _ in range(randrange(1,5))]
        sage: A = AdditiveAbelianGroup(ns)
        sage: from sage.groups.generic import has_order
        sage: el = A.random_element()
        sage: o = el.order()
        sage: has_order(el, o)
        True
        sage: has_order(el, o.factor())
        True
        sage: not_o = ZZ(randrange(100*o))
        sage: not_o += (not_o == o)
        sage: has_order(el, not_o)
        False

    Check for :issue:`37102`::

        sage: from sage.groups.generic import has_order
        sage: x = Mod(9, 24)
        sage: has_order(x, 0)
        False
        sage: has_order(x, -8)
        False

    .. NOTE::

        In some cases, order *testing* can be much faster than
        *computing* the order using :func:`order_from_multiple`.
    """
    if isinstance(n, sage.rings.integer.Integer):
        if n <= 0:
            return False
        n = n.factor()

    if operation in addition_names:
        isid = lambda el: not el
        mult = lambda el, n: multiple(el, n, operation='+')
    elif operation in multiplication_names:
        isid = lambda el: el.is_one()
        mult = multiple
    else:
        raise ValueError('unknown group operation')

    def _rec(Q, fn):
        if not fn:
            return isid(Q)

        if len(fn) == 1:
            p, k = fn[0]
            for _ in range(k):
                if isid(Q):
                    return False
                Q = mult(Q, p)
            return isid(Q)

        fl = fn[::2]
        fr = fn[1::2]
        l = prod(p**k for p, k in fl)
        r = prod(p**k for p, k in fr)
        L, R = mult(Q, r), mult(Q, l)
        return _rec(L, fl) and _rec(R, fr)

    return _rec(P, n)


def merge_points(P1, P2, operation='+',
                 identity=None, inverse=None, op=None, check=True):
    r"""
    Return a group element whose order is the lcm of the given elements.

    INPUT:

    - ``P1`` -- a pair `(g_1,n_1)` where `g_1` is a group element of order `n_1`
    - ``P2`` -- a pair `(g_2,n_2)` where `g_2` is a group element of order `n_2`
    - ``operation`` -- string; ``'+'`` (default) or ``'*'`` or other. If
      other, the following must be supplied:

      - ``identity`` -- the identity element for the group
      - ``inverse`` -- a function of one argument giving the inverse
        of a group element
      - ``op`` -- a function of 2 arguments defining the group
        binary operation

    OUTPUT: a pair `(g_3,n_3)` where `g_3` has order `n_3=\hbox{lcm}(n_1,n_2)`

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: from sage.groups.generic import merge_points
        sage: F.<a> = GF(3^6,'a')
        sage: b = a^7
        sage: c = a^13
        sage: ob = (3^6-1)//7
        sage: oc = (3^6-1)//13
        sage: merge_points((b,ob), (c,oc), operation='*')
        (a^4 + 2*a^3 + 2*a^2, 728)
        sage: d, od = merge_points((b,ob), (c,oc), operation='*')
        sage: od == d.multiplicative_order()
        True
        sage: od == lcm(ob, oc)
        True

        sage: # needs sage.libs.gap sage.rings.finite_rings sage.schemes
        sage: E = EllipticCurve([a^5 + 2*a^3 + 2*a^2 + 2*a, a^4 + a^3 + 2*a + 1])
        sage: P = E(2*a^5 + 2*a^4 + a^3 + 2, a^4 + a^3 + a^2 + 2*a + 2)
        sage: P.order()
        7
        sage: Q = E(2*a^5 + 2*a^4 + 1, a^5 + 2*a^3 + 2*a + 2)
        sage: Q.order()
        4
        sage: R, m = merge_points((P,7), (Q,4), operation='+')
        sage: R.order() == m
        True
        sage: m == lcm(7,4)
        True
    """
    from operator import mul, add

    g1, n1 = P1
    g2, n2 = P2

    if operation in multiplication_names:
        op = mul
        identity = g1.parent().one()
    elif operation in addition_names:
        op = add
        identity = g1.parent().zero()
    else:
        if op is None:
            raise ValueError("operation and identity must be specified")

    if check:
        if multiple(g1, n1, operation=operation) != identity or multiple(g2, n2, operation=operation) != identity:
            raise ValueError("the orders provided do not divide the orders of the points provided")

    # trivial cases
    if n1.divides(n2):
        return (g2, n2)
    if n2.divides(n1):
        return (g1, n1)

    m, k1, k2 = xlcm(n1, n2)
    m1 = n1 // k1
    m2 = n2 // k2
    g1 = multiple(g1, m1, operation=operation)
    g2 = multiple(g2, m2, operation=operation)
    return (op(g1, g2), m)


def structure_description(G, latex=False):
    r"""
    Return a string that tries to describe the structure of ``G``.

    This methods wraps GAP's ``StructureDescription`` method.

    For full details, including the form of the returned string and the
    algorithm to build it, see :gap:`GAP's documentation <chap39>`.

    INPUT:

    - ``latex`` -- boolean (default: ``False``); if ``True``, return a
      LaTeX formatted string

    OUTPUT: string

    .. WARNING::

        From GAP's documentation: The string returned by
        ``StructureDescription`` is **not** an isomorphism invariant:
        non-isomorphic groups can have the same string value, and two
        isomorphic groups in different representations can produce different
        strings.

    EXAMPLES::

        sage: # needs sage.groups
        sage: G = CyclicPermutationGroup(6)
        sage: G.structure_description()
        'C6'
        sage: G.structure_description(latex=True)
        'C_{6}'
        sage: G2 = G.direct_product(G, maps=False)
        sage: LatexExpr(G2.structure_description(latex=True))
        C_{6} \times C_{6}

    This method is mainly intended for small groups or groups with few
    normal subgroups. Even then there are some surprises::

        sage: D3 = DihedralGroup(3)                                                     # needs sage.groups
        sage: D3.structure_description()                                                # needs sage.groups
        'S3'

    We use the Sage notation for the degree of dihedral groups::

        sage: D4 = DihedralGroup(4)                                                     # needs sage.groups
        sage: D4.structure_description()                                                # needs sage.groups
        'D4'

    Works for finitely presented groups (:issue:`17573`)::

        sage: F.<x, y> = FreeGroup()                                                    # needs sage.groups
        sage: G = F / [x^2*y^-1, x^3*y^2, x*y*x^-1*y^-1]                                # needs sage.groups
        sage: G.structure_description()                                                 # needs sage.groups
        'C7'

    And matrix groups (:issue:`17573`)::

        sage: groups.matrix.GL(4,2).structure_description()                             # needs sage.libs.gap sage.modules
        'A8'
    """
    import re

    def correct_dihedral_degree(match):
        return "%sD%d" % (match.group(1), int(match.group(2)) // 2)

    description = str(G._gap_().StructureDescription())

    description = re.sub(r"(\A|\W)D(\d+)", correct_dihedral_degree, description)
    if not latex:
        return description
    description = description.replace("x", r"\times").replace(":", r"\rtimes")
    description = re.sub(r"([A-Za-z]+)([0-9]+)", r"\g<1>_{\g<2>}", description)
    description = re.sub(r"O([+-])", r"O^{\g<1>}", description)

    return description



# sage_setup: distribution = sagemath-categories
r"""
Multivariate Polynomial Rings over Generic Rings

Sage implements multivariate polynomial rings through several
backends. This generic implementation uses the classes ``PolyDict``
and ``ETuple`` to construct a dictionary with exponent tuples as keys
and coefficients as values.

AUTHORS:

- David Joyner and William Stein

- Kiran S. Kedlaya (2006-02-12): added Macaulay2 analogues of Singular
  features

- Martin Albrecht (2006-04-21): reorganize class hierarchy for singular
  rep

- Martin Albrecht (2007-04-20): reorganized class hierarchy to support
  Pyrex implementations

- Robert Bradshaw (2007-08-15): Coercions from rings in a subset of
  the variables.

EXAMPLES:

We construct the Frobenius morphism on `\GF{5}[x,y,z]` over
`\GF{5}`::

    sage: R.<x,y,z> = GF(5)[]
    sage: frob = R.hom([x^5, y^5, z^5])
    sage: frob(x^2 + 2*y - z^4)
    -z^20 + x^10 + 2*y^5
    sage: frob((x + 2*y)^3)                                                             # needs sage.rings.finite_rings
    x^15 + x^10*y^5 + 2*x^5*y^10 - 2*y^15
    sage: (x^5 + 2*y^5)^3                                                               # needs sage.rings.finite_rings
    x^15 + x^10*y^5 + 2*x^5*y^10 - 2*y^15

We make a polynomial ring in one variable over a polynomial ring in
two variables::

    sage: R.<x, y> = PolynomialRing(QQ, 2)
    sage: S.<t> = PowerSeriesRing(R)
    sage: t*(x+y)
    (x + y)*t

TESTS::

    sage: PolynomialRing(GF(5), 3, 'xyz').objgens()
    (Multivariate Polynomial Ring in x, y, z over Finite Field of size 5,
    (x, y, z))
"""
# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.ring import IntegralDomain
import sage.rings.fraction_field_element as fraction_field_element

from sage.rings.polynomial.multi_polynomial_ring_base import MPolynomialRing_base, is_MPolynomialRing
from sage.rings.polynomial.polynomial_singular_interface import PolynomialRing_singular_repr
from sage.rings.polynomial.polydict import PolyDict, ETuple
from sage.rings.polynomial.term_order import TermOrder

import sage.interfaces.abc

try:
    from sage.libs.pari.all import pari_gen
except ImportError:
    pari_gen = ()

from sage.structure.element import Element


class MPolynomialRing_macaulay2_repr:
    """
    A mixin class for polynomial rings that support conversion to Macaulay2.
    """
    def _macaulay2_init_(self, macaulay2=None):
        """
        EXAMPLES::

            sage: PolynomialRing(QQ, 'x', 2, implementation='generic')._macaulay2_init_()   # optional - macaulay2
            'sage...[symbol x0,symbol  x1, MonomialSize=>16, MonomialOrder=>GRevLex]'
        """
        if macaulay2 is None:
            from sage.interfaces.macaulay2 import macaulay2 as m2_default
            macaulay2 = m2_default
        return macaulay2._macaulay2_input_ring(self.base_ring(), self.gens(),
                                               self.term_order().macaulay2_str())


class MPolynomialRing_polydict(MPolynomialRing_macaulay2_repr, PolynomialRing_singular_repr, MPolynomialRing_base):
    """
    Multivariable polynomial ring.

    EXAMPLES::

        sage: R = PolynomialRing(Integers(12), 'x', 5); R
        Multivariate Polynomial Ring in x0, x1, x2, x3, x4 over Ring of integers modulo 12
        sage: loads(R.dumps()) == R
        True
    """
    from sage.rings.polynomial.multi_polynomial_element import MPolynomial_polydict as Element_hidden
    # should be just Element, once polynomial use new coercion framework

    def __init__(self, base_ring, n, names, order):
        from sage.rings.polynomial.polynomial_singular_interface import can_convert_to_singular
        order = TermOrder(order, n)
        # MPolynomialRing_base.__init__() normally initialises the base ring,
        # but it also needs the generators to construct a coercion map from the
        # base ring, and the base ring must be set to initialise the generators.
        # We set the base ring manually to break this circular dependency.
        self._base = base_ring
        # Construct the generators
        v = [0] * n
        one = base_ring.one()
        self._gens = []
        for i in range(n):
            v[i] = 1  # int's!
            self._gens.append(self.Element_hidden(self, {tuple(v): one}))
            v[i] = 0
        self._gens = tuple(self._gens)
        self._zero_tuple = tuple(v)
        MPolynomialRing_base.__init__(self, base_ring, n, names, order)
        self._has_singular = can_convert_to_singular(self)

    def _monomial_order_function(self):
        return self.__monomial_order_function

    def __eq__(self, other):
        """
        Check whether ``self`` is equal to ``other``.

        EXAMPLES::

            sage: R = PolynomialRing(Integers(10), 'x', 4)
            sage: loads(R.dumps()) == R
            True
        """
        if not isinstance(other, MPolynomialRing_base):
            return False
        return ((self.base_ring(), self.ngens(),
                self.variable_names(), self.term_order()) ==
                (other.base_ring(), other.ngens(),
                 other.variable_names(), other.term_order()))

    def __ne__(self, other):
        """
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: R = PolynomialRing(Integers(8), 'x', 3)
            sage: loads(R.dumps()) != R
            False
        """
        return not (self == other)

    def __hash__(self):
        """
        Compute the hash of ``self``.

        EXAMPLES::

            sage: h = hash(PolynomialRing(Integers(8), 'x', 3))
        """
        return hash((self.base_ring(), self.ngens(),
                     self.variable_names(), self.term_order()))

    def __call__(self, x=0, check=True):
        """
        Convert ``x`` to an element of this multivariate polynomial ring,
        possibly non-canonically.

        EXAMPLES:

        We create a Macaulay2 multivariate polynomial via ideal
        arithmetic, then convert it into R.

        ::

            sage: R.<x,y> = PolynomialRing(QQ, 2)
            sage: I = R.ideal([x^3 + y, y])
            sage: S = I._macaulay2_()                                    # optional - macaulay2
            sage: T = S*S*S                                              # optional - macaulay2
            sage: U = T.gens().entries().flatten()                       # optional - macaulay2
            sage: f = U[2]; f                                            # optional - macaulay2
             6      3 2    3
            x y + 2x y  + y
            sage: R(f.external_string())                                 # optional - macaulay2
            x^6*y + 2*x^3*y^2 + y^3

        Some other subtle conversions. We create polynomial rings in 2
        variables over the rationals, integers, and a finite field.

        ::

            sage: R.<x,y> = QQ[]
            sage: S.<x,y> = ZZ[]
            sage: T.<x,y> = GF(7)[]

        We convert from integer polynomials to rational polynomials,
        and back::

            sage: f = R(S.0^2 - 4*S.1^3); f
            -4*y^3 + x^2
            sage: parent(f)
            Multivariate Polynomial Ring in x, y over Rational Field
            sage: parent(S(f))
            Multivariate Polynomial Ring in x, y over Integer Ring

        We convert from polynomials over the finite field.

        ::

            sage: f = R(T.0^2 - 4*T.1^3); f
            3*y^3 + x^2
            sage: parent(f)
            Multivariate Polynomial Ring in x, y over Rational Field

        We dump and load the polynomial ring S::

            sage: S2 = loads(dumps(S))
            sage: S2 == S
            True

        Coerce works and gets the right parent.

        ::

            sage: parent(S2.coerce(S.0)) is S2
            True

        Conversion to reduce modulo a prime between rings with different
        variable names::

            sage: R.<x,y> = PolynomialRing(QQ,2)
            sage: S.<a,b> = PolynomialRing(GF(7),2)
            sage: f = x^2 + 2/3*y^3
            sage: S(f)
            3*b^3 + a^2

        Conversion from symbolic variables::

            sage: # needs sage.symbolic
            sage: x,y,z = var('x,y,z')
            sage: R = QQ['x,y,z']
            sage: type(x)
            <class 'sage.symbolic.expression.Expression'>
            sage: type(R(x))
            <class 'sage.rings.polynomial.multi_polynomial_libsingular.MPolynomial_libsingular'>
            sage: f = R(x^3 + y^3 - z^3); f
            x^3 + y^3 - z^3
            sage: type(f)
            <class 'sage.rings.polynomial.multi_polynomial_libsingular.MPolynomial_libsingular'>
            sage: parent(f)
            Multivariate Polynomial Ring in x, y, z over Rational Field

        A more complicated symbolic and computational mix. Behind the
        scenes Singular and Maxima are doing the real work.

        ::

            sage: # needs sage.symbolic
            sage: R = QQ['x,y,z']
            sage: f = (x^3 + y^3 - z^3)^10; f
            (x^3 + y^3 - z^3)^10
            sage: g = R(f); parent(g)
            Multivariate Polynomial Ring in x, y, z over Rational Field
            sage: (f - g).expand()
            0

        It intelligently handles conversions from polynomial rings in a subset
        of the variables too.

        ::

            sage: R = GF(5)['x,y,z']
            sage: S = ZZ['y']
            sage: R(7*S.0)
            2*y
            sage: T = ZZ['x,z']
            sage: R(2*T.0 + 6*T.1 + T.0*T.1^2)
            x*z^2 + 2*x + z

        ::

            sage: R = QQ['t,x,y,z']
            sage: S.<x> = ZZ['x']
            sage: T.<z> = S['z']
            sage: T
            Univariate Polynomial Ring in z over Univariate Polynomial Ring in x over Integer Ring
            sage: f = (x+3*z+5)^2; f
            9*z^2 + (6*x + 30)*z + x^2 + 10*x + 25
            sage: R(f)
            x^2 + 6*x*z + 9*z^2 + 10*x + 30*z + 25

        Arithmetic with a constant from a base ring::

            sage: R.<u,v> = QQ[]
            sage: S.<x,y> = R[]
            sage: u^3*x^2 + v*y
            u^3*x^2 + v*y

        Stacked polynomial rings convert into constants if possible. First,
        the univariate case::

            sage: R.<x> = QQ[]
            sage: S.<u,v> = R[]
            sage: S(u + 2)
            u + 2
            sage: S(u + 2).degree()
            1
            sage: S(x + 3)
            x + 3
            sage: S(x + 3).degree()
            0

        Second, the multivariate case::

            sage: R.<x,y> = QQ[]
            sage: S.<u,v> = R[]
            sage: S(x + 2*y)
            x + 2*y
            sage: S(u + 2*v)
            u + 2*v

        Conversion from strings::

            sage: R.<x,y> = QQ[]
            sage: R('x+(1/2)*y^2')
            1/2*y^2 + x
            sage: S.<u,v> = ZZ[]
            sage: S('u^2 + u*v + v^2')
            u^2 + u*v + v^2

        Foreign polynomial rings convert into the highest ring; the point
        here is that an element of T could convert to an element of R or an
        element of S; it is anticipated that an element of T is more likely
        to be "the right thing" and is historically consistent.

        ::

            sage: R.<x,y> = QQ[]
            sage: S.<u,v> = R[]
            sage: T.<a,b> = QQ[]
            sage: S(a + b)
            u + v

        TESTS:

        Check if we still allow nonsense (see :issue:`7951`)::

            sage: P = PolynomialRing(QQ, 0, '')
            sage: P('pi')                                                               # needs sage.symbolic
            Traceback (most recent call last):
            ...
            TypeError: unable to convert pi to a rational

        Check that it is possible to convert strings to iterated
        polynomial rings (see :issue:`13327`)::

            sage: Rm = QQ["a"]["b, c"]
            sage: Rm("a*b")
            a*b
            sage: parent(_) is Rm
            True

        Check that conversion from PARI works correctly (see
        :issue:`17974`)::

            sage: A.<a> = PolynomialRing(QQ)
            sage: B.<d,e> = PolynomialRing(A)
            sage: f = pari(a*d)                                                         # needs sage.libs.pari
            sage: B(f)
            a*d
            sage: f = pari(a*d - (a+1)*d*e^3 + a*d^2)                                   # needs sage.libs.pari
            sage: B(f)
            (-a - 1)*d*e^3 + a*d^2 + a*d

            sage: A.<a,b> = PolynomialRing(QQ)
            sage: B.<d,e> = PolynomialRing(A)
            sage: f = pari(a*d)                                                         # needs sage.libs.pari
            sage: B(f)
            a*d

        It is possible to convert `f` into `B` by using ``f.sage()``,
        but this requires specifying a ``locals`` argument::

            sage: f
            d*a
            sage: f.sage()
            Traceback (most recent call last):
            ...
            NameError: name 'd' is not defined
            sage: f.sage(locals={'a': a, 'd': d})
            a*d

        Check that :issue:`21999` is fixed::

            sage: R = QQbar['s,t']                                                      # needs sage.rings.number_field
            sage: type(R({(1,2): 3}).coefficients()[0])
            <class 'sage.rings.qqbar.AlgebraicNumber'>
        """
        from sage.rings.polynomial.multi_polynomial_element import MPolynomial_polydict
        import sage.rings.polynomial.polynomial_element as polynomial_element

        # handle constants that coerce into self.base_ring() first, if possible
        if isinstance(x, Element) and x.parent() is self.base_ring():
            # A Constant multi-polynomial
            return self({self._zero_tuple: x})

        try:
            y = self.base_ring().coerce(x)
            return MPolynomial_polydict(self, {self._zero_tuple: y})
        except TypeError:
            pass

        from .multi_polynomial import MPolynomial_libsingular

        if isinstance(x, MPolynomial_polydict):
            P = x.parent()

            if P is self:
                return x

            if P == self:
                return MPolynomial_polydict(self, x.element().dict())

            if self.base_ring().has_coerce_map_from(P):
                # it might be in the base ring (i.e. a poly ring over a poly ring)
                c = self.base_ring()(x)
                return MPolynomial_polydict(self, {self._zero_tuple: c})

            if len(P.variable_names()) == len(self.variable_names()):
                # Map the variables in some crazy way (but in order,
                # of course).  This is here since R(blah) is supposed
                # to be "make an element of R if at all possible with
                # no guarantees that this is mathematically solid."
                K = self.base_ring()
                D = x.element().dict()
                for i, a in D.items():
                    D[i] = K(a)
                return MPolynomial_polydict(self, D)

            if (set(P.variable_names()).issubset(set(self.variable_names()))
                and self.base_ring().has_coerce_map_from(P.base_ring())):
                # If the named variables are a superset of the input, map the variables by name
                return MPolynomial_polydict(self, self._extract_polydict(x))

            return MPolynomial_polydict(self,
                                        x._mpoly_dict_recursive(self.variable_names(),
                                                                self.base_ring()))

        elif isinstance(x, MPolynomial_libsingular):
            P = x.parent()
            if P == self:
                return MPolynomial_polydict(self, x.monomial_coefficients())

            if self.base_ring().has_coerce_map_from(P):
                # it might be in the base ring (i.e. a poly ring over a poly ring)
                c = self.base_ring()(x)
                return MPolynomial_polydict(self, {self._zero_tuple: c})

            if len(P.variable_names()) == len(self.variable_names()):
                # Map the variables in some crazy way (but in order,
                # of course).  This is here since R(blah) is supposed
                # to be "make an element of R if at all possible with
                # no guarantees that this is mathematically solid."
                K = self.base_ring()
                D = x.monomial_coefficients()
                for i, a in D.items():
                    D[i] = K(a)
                return MPolynomial_polydict(self, D)

            if (set(P.variable_names()).issubset(set(self.variable_names()))
                and self.base_ring().has_coerce_map_from(P.base_ring())):
                # If the named variables are a superset of the input, map the variables by name
                return MPolynomial_polydict(self, self._extract_polydict(x))

            return MPolynomial_polydict(self,
                                        x._mpoly_dict_recursive(self.variable_names(),
                                                                self.base_ring()))

        if isinstance(x, polynomial_element.Polynomial):
            return MPolynomial_polydict(self,
                                        x._mpoly_dict_recursive(self.variable_names(),
                                                                self.base_ring()))

        if isinstance(x, PolyDict):
            return MPolynomial_polydict(self, x)

        if isinstance(x, dict):
            K = self.base_ring()
            return MPolynomial_polydict(self, {i: K(a) for i, a in x.items()})

        if (isinstance(x, fraction_field_element.FractionFieldElement)
            and x.parent().ring() == self):
            if x.denominator() == 1:
                return x.numerator()

            raise TypeError("unable to coerce since the denominator is not 1")

        if isinstance(x, sage.interfaces.abc.SingularElement) and self._has_singular:
            self._singular_().set_ring()
            try:
                return x.sage_poly(self)
            except TypeError:
                raise TypeError("unable to coerce singular object")

        if hasattr(x, '_polynomial_'):
            return x._polynomial_(self)

        if isinstance(x, str):
            from sage.misc.sage_eval import sage_eval
            try:
                x = sage_eval(x, self.gens_dict_recursive())
            except NameError:
                raise TypeError("unable to evaluate {!r} in {}".format(x, self))
            return self(x)

        if isinstance(x, sage.interfaces.abc.Macaulay2Element):
            try:
                s = x.sage_polystring()
                if len(s) == 0:
                    raise TypeError
                # NOTE: It's CRUCIAL to use the eval command as follows,
                # i.e., with the gen dict as the third arg and the second
                # empty.  Otherwise pickling won't work after calls to this eval!!!
                # This took a while to figure out!
                return self(eval(s, {}, self.gens_dict()))
            except (AttributeError, TypeError, NameError, SyntaxError):
                raise TypeError("Unable to coerce macaulay2 object")
            return MPolynomial_polydict(self, x)

        if isinstance(x, pari_gen) and x.type() == 't_POL':
            # This recursive approach is needed because PARI
            # represents multivariate polynomials as iterated
            # univariate polynomials.  Below, v is the variable
            # with highest priority, and the x[i] are expressions
            # in the remaining variables.
            d = x.poldegree()
            if d.type() == 't_INFINITY':
                return self.zero()
            v = self.gens_dict_recursive()[str(x.variable())]
            return sum(self(x[i]) * v ** i for i in range(d + 1))

        if isinstance(x, dict):
            return MPolynomial_polydict(self, x)

        c = self.base_ring()(x)
        return MPolynomial_polydict(self, {self._zero_tuple: c})

# The following methods are handy for implementing Groebner
# basis algorithms. They do only superficial type/sanity checks
# and should be called carefully.

    def monomial_quotient(self, f, g, coeff=False):
        r"""
        Return ``f/g``, where both ``f`` and ``g`` are treated as monomials.

        Coefficients are ignored by default.

        INPUT:

        - ``f`` -- monomial

        - ``g`` -- monomial

        - ``coeff`` -- divide coefficients as well (default: ``False``)

        OUTPUT: monomial

        EXAMPLES::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(QQ, 3, order='degrevlex')
            sage: P.monomial_quotient(3/2*x*y, x)
            y

        ::

            sage: P.monomial_quotient(3/2*x*y, 2*x, coeff=True)
            3/4*y

        TESTS::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: R.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.monomial_quotient(x*y, x)
            y

        ::

            sage: P.monomial_quotient(x*y, R.gen())
            y

        ::

            sage: P.monomial_quotient(P(0), P(1))
            0

        ::

            sage: P.monomial_quotient(P(1), P(0))
            Traceback (most recent call last):
            ...
            ZeroDivisionError

        ::

            sage: P.monomial_quotient(P(3/2), P(2/3), coeff=True)
            9/4

        ::

            sage: P.monomial_quotient(x, y) # Note the wrong result
            x*y^-1

        ::

            sage: P.monomial_quotient(x, P(1))
            x

        .. NOTE::

            Assumes that the head term of f is a multiple of the head
            term of g and return the multiplicant m. If this rule is
            violated, funny things may happen.
        """
        from sage.rings.polynomial.multi_polynomial_element import MPolynomial_polydict

        if not f:
            return f
        if not g:
            raise ZeroDivisionError

        fd = f.monomial_coefficients()
        gd = g.monomial_coefficients()

        if not coeff:
            f = next(iter(fd))
            g = next(iter(gd))
            coeff = self.base_ring().one()
        else:
            f, cf = next(iter(fd.items()))
            g, cg = next(iter(gd.items()))
            coeff = self.base_ring()(cf / cg)

        res = f.esub(g)

        return MPolynomial_polydict(self, PolyDict({res: coeff}))

    def monomial_lcm(self, f, g):
        """
        LCM for monomials. Coefficients are ignored.

        INPUT:

        - ``f`` -- monomial

        - ``g`` -- monomial

        OUTPUT: monomial

        EXAMPLES::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.monomial_lcm(3/2*x*y, x)
            x*y

        TESTS::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: R.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.monomial_lcm(x*y, R.gen())
            x*y

        ::

            sage: P.monomial_lcm(P(3/2), P(2/3))
            1

        ::

            sage: P.monomial_lcm(x, P(1))
            x
        """
        one = self.base_ring().one()

        f, = f.monomial_coefficients()
        g, = g.monomial_coefficients()

        length = len(f)

        res = {i: max(f[i], g[i])
               for i in f.common_nonzero_positions(g)}

        return self(PolyDict({ETuple(res, length): one}))

    def monomial_reduce(self, f, G):
        r"""
        Try to find a ``g`` in ``G`` where ``g.lm()`` divides ``f``.

        If found, ``(flt,g)`` is returned, ``(0,0)`` otherwise, where
        ``flt`` is ``f/g.lm()``. It is assumed that ``G`` is iterable and contains
        ONLY elements in this ring.

        INPUT:

        - ``f`` -- monomial

        - ``G`` -- list/set of mpolynomials

        EXAMPLES::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z>=MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: f = x*y^2
            sage: G = [3/2*x^3 + y^2 + 1/2, 1/4*x*y + 2/7, P(1/2)]
            sage: P.monomial_reduce(f,G)
            (y, 1/4*x*y + 2/7)

        ::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(Zmod(23432),3, order='degrevlex')
            sage: f = x*y^2
            sage: G = [3*x^3 + y^2 + 2, 4*x*y + 7, P(2)]
            sage: P.monomial_reduce(f,G)
            (y, 4*x*y + 7)

        TESTS::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z>=MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: f = x*y^2
            sage: G = [3/2*x^3 + y^2 + 1/2, 1/4*x*y + 2/7, P(1/2)]

        ::

            sage: P.monomial_reduce(P(0),G)
            (0, 0)

        ::

            sage: P.monomial_reduce(f,[P(0)])
            (0, 0)
        """
        if not f:
            return 0, 0
        for g in G:
            t = g.lm()
            try:
                if self.monomial_divides(t, f):
                    return self.monomial_quotient(f, t), g
            except ZeroDivisionError:
                return 0, 0
        return 0, 0

    def monomial_divides(self, a, b):
        """
        Return ``False`` if ``a`` does not divide ``b`` and ``True`` otherwise.

        INPUT:

        - ``a`` -- monomial

        - ``b`` -- monomial

        OUTPUT: boolean

        EXAMPLES::

            sage: P.<x,y,z> = PolynomialRing(ZZ,3, order='degrevlex')
            sage: P.monomial_divides(x*y*z, x^3*y^2*z^4)
            True
            sage: P.monomial_divides(x^3*y^2*z^4, x*y*z)
            False

        TESTS::

            sage: P.<x,y,z> = PolynomialRing(ZZ,3, order='degrevlex')
            sage: P.monomial_divides(P(1), P(0))
            True
            sage: P.monomial_divides(P(1), x)
            True
        """
        if not b:
            return True
        if not a:
            raise ZeroDivisionError

        a, = a.monomial_coefficients()
        b, = b.monomial_coefficients()

        return all(b[i] >= a[i]
                   for i in b.common_nonzero_positions(a))

    def monomial_pairwise_prime(self, h, g):
        r"""
        Return ``True`` if ``h`` and ``g`` are pairwise prime.

        Both are treated as monomials.

        INPUT:

        - ``h`` -- monomial

        - ``g`` -- monomial

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.monomial_pairwise_prime(x^2*z^3, y^4)
            True

        ::

            sage: P.monomial_pairwise_prime(1/2*x^3*y^2, 3/4*y^3)
            False

        TESTS::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: Q.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.monomial_pairwise_prime(x^2*z^3, Q('y^4'))
            True

        ::

            sage: P.monomial_pairwise_prime(1/2*x^3*y^2, Q(0))
            True

        ::

            sage: P.monomial_pairwise_prime(P(1/2),x)
            False
        """
        not_g = not g
        not_h = not h
        if not_g and not_h:  # GCD(0,0) = 0
            return False
        if not_g or not_h:  # GCD(x,0) = 1
            return True
        return self.monomial_lcm(g, h) == g * h

    def monomial_all_divisors(self, t):
        r"""
        Return a list of all monomials that divide ``t``, coefficients are
        ignored.

        INPUT:

        - ``t`` -- a monomial

        OUTPUT: list of monomials

        EXAMPLES::

            sage: from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict_domain
            sage: P.<x,y,z> = MPolynomialRing_polydict_domain(QQ,3, order='degrevlex')
            sage: P.monomial_all_divisors(x^2*z^3)
            [x, x^2, z, x*z, x^2*z, z^2, x*z^2, x^2*z^2, z^3, x*z^3, x^2*z^3]

        ALGORITHM: addwithcarry idea by Toon Segers
        """

        def addwithcarry(tempvector, maxvector, pos):
            if tempvector[pos] < maxvector[pos]:
                tempvector[pos] += 1
            else:
                tempvector[pos] = 0
                tempvector = addwithcarry(tempvector, maxvector, pos + 1)
            return tempvector

        if not t.is_monomial():
            raise TypeError("only monomials are supported")

        R = self
        one = self.base_ring().one()
        M = list()

        v, = t.monomial_coefficients()
        maxvector = list(v)

        tempvector = [0] * len(maxvector)

        pos = 0

        while tempvector != maxvector:
            tempvector = addwithcarry(list(tempvector), maxvector, pos)
            M.append(R(PolyDict({ETuple(tempvector): one})))
        return M

    def sum(self, terms):
        r"""
        Return a sum of elements of this multipolynomial ring.

        This is method is much faster than the Python builtin :func:`sum`.

        EXAMPLES::

            sage: R = QQ['x']
            sage: S = R['y, z']
            sage: x = R.gen()
            sage: y, z = S.gens()
            sage: S.sum([x*y, 2*x^2*z - 2*x*y, 1 + y + z])
            (-x + 1)*y + (2*x^2 + 1)*z + 1

        Comparison with builtin :func:`sum`::

            sage: sum([x*y, 2*x^2*z - 2*x*y, 1 + y + z])
            (-x + 1)*y + (2*x^2 + 1)*z + 1
        """
        elt = PolyDict({}, check=False)
        for t in terms:
            elt += self(t).element()
        # NOTE: here we should be using self.element_class but polynomial rings are not complient
        # with categories...
        from sage.rings.polynomial.multi_polynomial_element import MPolynomial_polydict
        return MPolynomial_polydict(self, elt)


class MPolynomialRing_polydict_domain(IntegralDomain,
                                      MPolynomialRing_polydict):
    def __init__(self, base_ring, n, names, order):
        order = TermOrder(order, n)
        MPolynomialRing_polydict.__init__(self, base_ring, n, names, order)

    def is_integral_domain(self, proof=True):
        return True

    def is_field(self, proof=True):
        if self.ngens() == 0:
            return self.base_ring().is_field(proof)
        return False
